"""Authorization ruleset handling.

The :class:`UserScopes` class handles whether a dataset, table or field can be accessed.
The other classes in this module ease to retrieval of permission objects.
"""
from __future__ import annotations

from functools import wraps
from typing import Callable, Dict, Iterable, List, Optional

import methodtools

from schematools.types import (
    DatasetFieldSchema,
    DatasetSchema,
    DatasetTableSchema,
    Permission,
    PermissionLevel,
    ProfileDatasetSchema,
    ProfileSchema,
    ProfileTableSchema,
)

__all__ = ("UserScopes",)


def abort_on_highest(func: Callable[..., Permission]):
    """Decorator to abort searching for permissions when the highest value was found.

    This is an internal helper to make the permission search code DRY.
    """

    @wraps(func)
    def _abort_on_highest(*args, **kwargs) -> Permission:
        try:
            return func(*args, **kwargs)
        except HighestPermissionFound as e:
            return e.permission

    return _abort_on_highest


class HighestPermissionFound(Exception):
    def __init__(self, permission: Permission):
        self.permission = permission


class PermissionCollection:
    """A helper class to ease collecting permissions from profile objects.

    This reduces the DRY code in permission searches, when used together with ``abort_on_highest``.
    It helps to find the highest permission level,
    and automatically aborts when the highest possible value is found.
    """

    def __init__(self):
        self._collection = []

    def append(self, permission: Permission):
        """Add a new discovered permission level to this collection."""
        self._collection.append(permission)
        if permission.level is PermissionLevel.highest:
            # Avoid further code flow.
            raise HighestPermissionFound(permission)

    @property
    def highest_value(self) -> Permission:
        """Return the highest found permission."""
        return max(self._collection, default=Permission.none)

    def __bool__(self):
        return bool(self._collection)


class UserScopes:
    """A request-like object that tells what the current user may access.

    A UserScopes encapsulates the scopes on a request and the profiles that apply,
    and performs permission checks against a schema.

    All ``has_...()`` functions are used for permission checks.
    Internally, these read the schema and profile data for the authorization matrix.

    * By default, all fields can be read unless the schema defines an "auth" field.
    * The "auth" flags in schema files act as a blacklist: no access, except for some roles.
    * The "profile" rules open up certain fields, hence whitelist features.
    """

    def __init__(
        self,
        query_params: Dict[str, ...],
        request_scopes: Iterable[str],
        all_profiles: Optional[Iterable[ProfileSchema]] = None,
    ):
        """Initialize the user scopes object.

        Args:
            query_params: The search query filter (e.g. request.GET).
            request_scopes: The scopes granted to a request.
            all_profiles: All profiles that need to be loaded.
                If not None, this iterable is stored and converted to list
                the first time it is needed.
        """
        self._query_param_names = [param for param, value in query_params.items() if value]
        self._all_profiles = all_profiles
        self._scopes = frozenset(request_scopes)

    def add_query_params(self, params: List[str]):
        """Tell that the request has extra (implicit) parameters that are satisfied.

        For example, the detail URL of a resource already implicitly passes the
        identifier of a resource. Hence, this parameter no longer needs to be given
        found in any additional search filters or query string.
        """
        self._query_param_names.extend(params)

    @methodtools.lru_cache()  # type: ignore[misc]
    def has_all_scopes(self, *needed_scopes: str) -> bool:
        """Check whether the request has all scopes.

        This performs an AND check: all scopes should be present.
        """
        return self._scopes.issuperset(needed_scopes)

    @methodtools.lru_cache()  # type: ignore[misc]
    def has_any_scope(self, *needed_scopes: str) -> bool:
        """Check whether the request grants one of the given scopes.

        This performs an OR check: having one of the scopes gives access.
        """
        needed_scopes = set(needed_scopes)
        return not needed_scopes or any(scope in needed_scopes for scope in self._scopes)

    def has_dataset_access(self, dataset: DatasetSchema) -> Permission:
        """Tell whether a dataset can be accessed."""
        return self._has_dataset_auth_access(dataset) or self._has_dataset_profile_access(
            dataset.id
        )

    def has_table_access(self, table: DatasetTableSchema) -> Permission:
        """Tell whether a table can be accessed, and return the permission level."""
        # When the user has an "auth" scope, they may always enter.
        # Otherwise, the user can only enter when the required profile rules are satisfied,
        # which includes mandatory filtersets.
        return self._has_table_auth_access(table) or self._has_table_profile_access(table)

    def has_field_access(self, field: DatasetFieldSchema) -> Permission:
        """Tell whether a field may be read."""
        # Again, when a field "auth" scope is satisfied, no further checks are done.
        # Otherwise, the field + table rules are checked from the profile.
        return self._has_field_auth_access(field) or self._has_field_profile_access(field)

    def _has_dataset_auth_access(self, dataset: DatasetSchema) -> Permission:
        """Tell whether the 'auth' rules give access to the dataset."""
        if self.has_any_scope(*dataset.auth):
            return Permission(PermissionLevel.highest, source="dataset.auth")
        else:
            return Permission.none

    def _has_table_auth_access(self, table: DatasetTableSchema) -> Permission:
        """Tell whether the 'auth' rules give access to the table."""
        if self.has_any_scope(*table.auth) and self.has_any_scope(*table.dataset.auth):
            return Permission(
                PermissionLevel.highest, source="table.auth" if table.auth else "dataset.auth"
            )
        else:
            return Permission.none

    def _has_field_auth_access(self, field: DatasetFieldSchema) -> Permission:
        """Tell whether the 'auth' rules give access to the table."""
        if (
            self.has_any_scope(*field.auth)
            and self.has_any_scope(*field.table.auth)
            and self.has_any_scope(*field.table.dataset.auth)
        ):
            return Permission(
                PermissionLevel.highest,
                source=(
                    "field.auth"
                    if field.auth
                    else ("table.auth" if field.table.auth else "dataset.auth")
                ),
            )
        else:
            return Permission.none

    @methodtools.lru_cache()
    def _has_dataset_profile_access(self, dataset_id: str) -> Permission:
        """Give the permission access level for a dataset, as defined by the profile."""
        return max(
            (
                profile_dataset.permissions
                for profile_dataset in self.get_active_profile_datasets(dataset_id)
            ),
            default=Permission.none,
        )

    @methodtools.lru_cache()
    @abort_on_highest
    def _has_table_profile_access(self, table: DatasetTableSchema) -> Permission:
        """Give the permission level for a table.

        When a dataset defines global permissions without explicitly mentioning the table,
        these permissions are "inherited" and used.
        """
        dataset_id = table.dataset.id
        table_id = table.id
        permissions = PermissionCollection()

        for profile_dataset in self.get_active_profile_datasets(dataset_id):
            # If a profile defines "read" on the whole dataset, without explicitly
            # mentioning a table, this means the table can also be read unconditionally.
            profile_table = profile_dataset.tables.get(table_id, None)
            if profile_table is None:
                if dataset_permission := profile_dataset.permissions:
                    permissions.append(dataset_permission)

            # Otherwise see if the table can be included (mandatory filters match)
            elif self._may_include_profile_table(profile_table):
                permissions.append(profile_table.permissions)

        # Datasets may a permission that also covers this table,
        # but tables could also define an explicit permission. See who wins.
        return permissions.highest_value

    @abort_on_highest
    def _has_field_profile_access(self, field: DatasetFieldSchema) -> Permission:
        """Give the permission level for a field based on a profile.

        Fields have a special case: if a specific permission is defined, use that.
        This may "limit" the actual permission. For example, the table gives "read" permission,
        but the field may state "encoded" as the level. Since a default is defined for the field,
        that's being used.
        """
        permissions = PermissionCollection()
        field_id = field.id
        table_id = field.table.id

        # First see if there is an explicit definition for a field:
        for profile_dataset in self.get_active_profile_datasets(field.table.dataset.id):
            # If a profile defines "read" on the whole dataset, without explicitly
            # mentioning the table, this means the table can also be read unconditionally.
            profile_table = profile_dataset.tables.get(table_id, None)
            if profile_table is None:
                if dataset_permission := profile_dataset.permissions:
                    permissions.append(dataset_permission)
                continue

            # See if the table can be included (mandatory filters match)
            if not self._may_include_profile_table(profile_table):
                continue

            # See if the table defines the current field
            try:
                field_permission = profile_table.fields[field_id]
            except KeyError:
                # No explicit field defined, consider table as fallback
                # Some tables define global permissions without mentioning the field.
                # These get preference over explicit field permissions from other profiles.
                # When we only find profiles that mention the field explicitly,
                # that highest level will be returned instead.
                table_permission = profile_table.permissions
                if table_permission and table_permission.level > PermissionLevel.SUBOBJECTS_ONLY:
                    permissions.append(table_permission)
                continue

            permissions.append(field_permission)

        return permissions.highest_value

    @methodtools.lru_cache()
    def get_active_profile_datasets(self, dataset_id: str) -> List[ProfileDatasetSchema]:
        """Find all profiles that mention a dataset and match the scopes.

        This already checks whether the mandatory user scopes are set.
        """
        if self._all_profiles is None:
            self._all_profiles = []
        elif not isinstance(self._all_profiles, list):
            self._all_profiles = list(self._all_profiles)  # perform query on demand.

        return [
            profile_dataset
            for profile in self._all_profiles
            # Profiles are only activated when:
            # - ALL scopes are matched
            # - dataset is mentioned in the profile
            if self.has_all_scopes(*profile.scopes)
            and (profile_dataset := profile.datasets.get(dataset_id)) is not None
        ]

    @methodtools.lru_cache()
    def get_active_profile_tables(
        self, dataset_id: str, table_id: str
    ) -> List[ProfileTableSchema]:
        """Find all profiles that mention a particular table and give access.

        This already checks whether the table passes the `mandatoryFilterSets` check,
        and whether the scopes of the dataset match.
        """
        return [
            profile_table
            for profile_dataset in self.get_active_profile_datasets(dataset_id)
            # Profiles are only activated when:
            # - table is mentioned in the profile
            # - ALL scopes are matched (tested for dataset already)
            # - ONE mandatory filter is matched (if filters are required)
            if (profile_table := profile_dataset.tables.get(table_id)) is not None
            and self._may_include_profile_table(profile_table)
        ]

    def _may_include_profile_table(self, profile_table: ProfileTableSchema):
        """Check whether the table rules are applicable to the current user.

        This checks whether any of the mandatory filtersets from a ProfileTableSchema were queried.
        """
        mandatory_filtersets = profile_table.mandatory_filtersets

        # Table is OK when:
        # - there are no mandatory filter sets
        # - one of the rules is matched.
        return not mandatory_filtersets or any(
            _match_filter_rule(rule, self._query_param_names) for rule in mandatory_filtersets
        )

    def __str__(self) -> str:
        """str implementation, for debugging purposes."""
        scopes = ", ".join(repr(scope) for scope in self._scopes)
        return f"UserScopes({scopes})"


def _match_filter_rule(rule: Iterable[str], query_param_names: Iterable[str]) -> bool:
    """Tell whether a mandatory filter rule is matched.

    This happens when ALL required filters are present in the query string.
    """
    return all(filter_name in query_param_names for filter_name in rule)



from typing import Final

RELATION_INDICATOR: Final[str] = "_"
MAX_TABLE_NAME_LENGTH: Final[int] = 63  # limitation of PostgreSQL
TMP_TABLE_POSTFIX: Final[str] = "_new"
TABLE_INDEX_POSTFIX: Final[str] = "_idx"



from __future__ import annotations

import os
from typing import TYPE_CHECKING, Optional

from simple_singleton import Singleton

if TYPE_CHECKING:
    from schematools.types import DatasetSchema


class DatasetCollection(metaclass=Singleton):
    """Holding class for a collection of datasets.

    This can hold a cache of datasets that have already been collected.
    This class is a singleton class, so every DatasetSchema can have
    a reference to it, without creating redundancy.
    """

    def __init__(self) -> None:
        self.datasets_cache = {}

    def _load_dataset(self, dataset_id: str) -> Optional[DatasetSchema]:
        """Loads the dataset from SCHEMA_URL.

        If SCHEMA_URL is not defined, return None.
        """
        # Avoid circular import problem.
        from schematools.utils import dataset_schema_from_url

        try:
            schemas_url = os.environ["SCHEMA_URL"]
        except KeyError:
            return None

        return dataset_schema_from_url(schemas_url, dataset_id, prefetch_related=True)

    def add_dataset(self, dataset: DatasetSchema) -> None:
        self.datasets_cache[dataset.id] = dataset

    def get_dataset(self, dataset_id: str) -> DatasetSchema:
        """Gets a dataset by id from the cache.

        If not available, load the dataset from the SCHEMA_URL location.
        NB. Because dataset schemas are imported into the Postgresql database
        by the DSO API, there is a chance that the dataset that is loaded from SCHEMA_URL
        differs from the definition that is in de Postgresql database.
        """
        try:
            return self.datasets_cache[dataset_id]
        except KeyError:
            dataset = self._load_dataset(dataset_id)
            if dataset is None:
                raise ValueError(f"Dataset {dataset_id} is missing.") from None
            self.add_dataset(dataset)
            return dataset



"""Converting a GeoJSON input file to Amsterdam Schema"""
import re
from copy import deepcopy
from decimal import Decimal as D
from os.path import basename, splitext
from typing import Iterable, List, Optional, Tuple

from schematools.importer.geojson import read_geojson, split_id
from schematools.introspect.utils import DATASET_TMPL, TABLE_TMPL

ID_FORMAT = re.compile(r"^([a-z0-9_]+)[/.](\d+)$", re.I)


def introspect_geojson_files(dataset_id: str, files: List[str]) -> dict:
    """Generate Amsterdam Schema from GeoJSON files."""
    tables = []
    for file in files:
        tables.extend(introspect_geojson_file(file))

    aschema = deepcopy(DATASET_TMPL)
    aschema["id"] = dataset_id
    aschema["tables"] = tables
    return aschema


def introspect_geojson_file(file_name) -> List[dict]:
    """Convert a single GeoJSON file into a JSON Schema"""
    features = read_geojson(file_name)
    return geojson_to_table(features, file_name=file_name)


def geojson_to_table(geojson_features: Iterable[dict], file_name: str) -> List[dict]:
    """Read the GeoJSON contents, return the table with JSON Schema.

    :param filename: This is provided for error messages.
    """
    default_name = splitext(basename(file_name))[0]
    all_schema = {}
    all_properties = {}

    for feature in geojson_features:
        # Parse feature['id'], determine the datatype of this feature
        table_name, id_value = _parse_id(feature, default_name)

        # These values are filled by reference, linked in all_*
        schema = all_schema.setdefault(table_name, {})
        properties = all_properties.setdefault(table_name, {})

        # Fill the "properties" section of the table schema
        _fill_properties(feature, properties, id_value)

        # Auto-detect display field
        display_field = _get_display(feature)
        if display_field is not None:
            schema["display"] = display_field

        # When all types are determined, the loop can be exited.
        # If there were None values, the next record is examined.
        if all(properties.values()):
            break

    # Overlay all results into the amsterdam schem table format.
    result = []
    for name in all_schema:
        table = deepcopy(TABLE_TMPL)
        table["id"] = name if name == default_name else f"{default_name}_{name}"
        table["schema"].update(all_schema[name])
        table["schema"]["properties"].update(all_properties[name])
        result.append(table)

    return result


def _parse_id(feature, default_name) -> Tuple[str, Optional[str]]:
    """Support datatype/PKVALUE as id value"""
    # Support optional "id" field at feature level
    try:
        id_value = feature["id"]
        return split_id(id_value)
    except (KeyError, ValueError):
        return default_name, None


def _fill_properties(feature: dict, properties: dict, id_value=None):
    """Fill the "properties" section of the table schema"""
    # Introspect "geometry" field at feature level
    if id_value is not None:
        properties["id"] = _build_geojson_field(id_value)

    geom_type = feature["geometry"]["type"]
    properties["geometry"] = {"$ref": f"https://geojson.org/schema/{geom_type}.json"}

    # Introspect remaining properties block
    for name, value in feature.get("properties", {}).items():
        if name[0] in "@$":
            name = name[1:]

        column = _build_geojson_field(value)
        if column is None and name in properties:
            # Don't override existing value with none
            continue

        properties[name] = column


def _get_display(feature: dict) -> Optional[str]:
    """Auto-detect some fields as possible 'display' field"""
    properties = feature.get("properties", {})
    for name in ("name", "title"):
        if name in properties:
            return name

    return None


def _build_geojson_field(value):
    """Determine the table column based on a value."""
    if value is None:
        return None
    elif isinstance(value, bool):
        return {"type": "boolean"}
    elif isinstance(value, (int, float, D)):
        return {"type": "number"}
    else:
        return {"type": "string"}



from django.db.models import QuerySet


class DatasetQuerySet(QuerySet):
    """Extra ORM methods for the Dataset model."""

    def db_enabled(self):
        """Return all datasets for which models should be created"""
        return self.filter(enable_db=True)

    def endpoint_enabled(self):
        """Return the datasets for which"""
        return self.filter(enable_db=False, endpoint_url__isnull=False)

    def api_enabled(self):
        return self.filter(enable_api=True)



import json
from pathlib import PosixPath
from typing import Any, Callable, Dict, Iterator, List, Optional, cast

import ndjson
from shapely.geometry import shape

from schematools import MAX_TABLE_NAME_LENGTH, RELATION_INDICATOR
from schematools.types import DatasetTableSchema
from schematools.utils import to_snake_case

from . import get_table_name
from .base import BaseImporter, Row


class NDJSONImporter(BaseImporter):
    """Import an NDJSON file into the database."""

    def _get_through_fields_mapper(
        self, dataset_table: DatasetTableSchema
    ) -> Optional[Callable[[Row], Row]]:
        """Maps fields in ndjson to proper fieldnames and structure.

        When through tables (1-N and NM relations) are imported directly
        from the GOB graphql API, the fieldnames in the ndjson are not
        following the names of the associated amsterdam schema.

        The importer is built around the amsterdam schema fieldnames.
        So, incoming ndjson records needs to be mapped to the appropriate
        amsterdam schema field names, to be able to import the ndjson
        records correctly.

        """
        field_mapping = {}
        # This is mapping the GOBModel `entity_id` to `Id`
        # that needs to be used in GraphQL queries.
        id_name_mapping = {
            "identificatie": "Id",
            "volgnummer": "Volgnummer",
            "dossier": "Id",
            "documentnummer": "Id",
            "vestigingsnummer": "Id",
            "sbiActiviteitNummer": "Id",
            "kvknummer": "Id",
            "wozdeelobjectnummer": "Id",
            "wozobjectnummer": "Id",
            "id": "Id",
            "code": "Id",
        }
        through_field_ids = dataset_table.data.get("throughFields")
        if through_field_ids is None:
            return None

        for direction_prefix, field_id in zip(("src", "dst"), through_field_ids):
            field = dataset_table.get_field_by_id(field_id)
            if field is None:
                self.logger.log_warning("No through field found: %s", field_id)
                return None

            related_table = field.related_table
            if related_table is None:
                self.logger.log_warning("No related_table found for: %s", field_id)
                return None

            field_mapping[field_id] = [
                (idf, f"{direction_prefix}{id_name_mapping[idf]}")
                for idf in related_table.identifier
            ]

        def _map_fields(row: Row) -> Row:
            for field_id, idfs__in_names in field_mapping.items():
                if len(idfs__in_names) > 1:  # compound key
                    row[field_id] = {idf: row[in_name] for idf, in_name in idfs__in_names}
                else:
                    row[field_id] = row[idfs__in_names[0][1]]
            return row

        return _map_fields

    def parse_records(  # type: ignore[override]
        self,
        file_name: PosixPath,
        dataset_table: DatasetTableSchema,
        db_table_name: Optional[str] = None,
        is_through_table: bool = False,
        **kwargs: Any,
    ) -> Iterator[Dict[str, List[Row]]]:
        """Provide an iterator the reads the NDJSON records."""
        fields_provenances = kwargs.pop("fields_provenances", {})
        identifier = dataset_table.identifier
        has_compound_key = dataset_table.has_compound_key
        through_fields_mapper = None
        if is_through_table:
            through_fields_mapper = self._get_through_fields_mapper(dataset_table)

        if db_table_name is None:
            db_table_name = get_table_name(dataset_table)

        table_name = dataset_table.name

        # Set up info for the special-case fields
        relation_field_info = []
        nm_relation_field_info = []
        nested_field_info = []
        inactive_relation_info = []
        jsonpath_provenance_info = []
        geo_fields = []

        for field in dataset_table.fields:
            # XXX maybe this is too much of a dirty hack and it would be better
            # to have an external configuration that determines which fields should
            # be flattened to strings
            comment = field.get("$comment")
            if comment is not None and "*stringify*" in comment:
                inactive_relation_info.append(field)
            if field.relation is not None:
                relation_field_info.append(field)
            field_provenance = field.provenance
            if field_provenance is not None and field_provenance.startswith("$"):
                jsonpath_provenance_info.append(field.name)
            if field.is_geo:
                geo_fields.append(field.name)
            if field.is_through_table and field.nm_relation is not None:
                nm_relation_field_info.append(field)
            if field.is_nested_table:
                nested_field_info.append(field)

        with open(file_name) as fh:
            for _row in ndjson.reader(fh):
                if through_fields_mapper is not None:
                    _row = through_fields_mapper(_row)
                row = Row(_row, fields_provenances=fields_provenances)
                for ir_field in inactive_relation_info:
                    row[ir_field.name] = json.dumps(row[ir_field.id])
                for field_name in jsonpath_provenance_info:
                    row[field_name] = row[field_name]  # uses Row to get from object
                sub_rows = {}
                for field_name in geo_fields:
                    geo_value = row[field_name]
                    if geo_value is not None:
                        wkt = shape(geo_value).wkt
                        row[field_name] = f"SRID={self.srid};{wkt}"

                if not dataset_table.is_autoincrement:
                    id_value = ".".join(str(row[fn]) for fn in identifier)
                    if has_compound_key:
                        row["id"] = id_value

                for rel_field in relation_field_info:
                    relation_field_name = rel_field.name
                    # Only process relation if data is available in incoming row
                    if rel_field.id not in row:
                        continue
                    relation_field_value = row[rel_field.id]
                    if rel_field.is_object:
                        fk_value_parts = []
                        for sub_field in rel_field.get_sub_fields(add_prefixes=True):
                            # Ignore temporal fields
                            if sub_field.is_temporal:
                                continue

                            sub_field_id = sub_field.id.rsplit(RELATION_INDICATOR, 1)[1]
                            if relation_field_value is None:
                                sub_field_value = None
                            else:
                                sub_field_value = relation_field_value[sub_field_id]
                                fk_value_parts.append(sub_field_value)
                            row[sub_field.name] = sub_field_value
                        # empty fk_value_parts should result in None value
                        relation_field_value = ".".join([str(p) for p in fk_value_parts]) or None
                    row[f"{relation_field_name}_id"] = relation_field_value
                    del row[rel_field.id]

                for n_field in nested_field_info:
                    field_name = to_snake_case(n_field.name)
                    nested_row_records = []

                    if not row[n_field.id]:
                        continue
                    for nested_row in row[n_field.id]:
                        # When the identifier is compound, we can assume
                        # that an extra 'id' field will be available, because
                        # Django cannot live without it.
                        id_fields = dataset_table.identifier
                        id_field_name = "id" if len(id_fields) > 1 else id_fields[0]
                        nested_row_record = {}
                        nested_row_record["parent_id"] = row[id_field_name]
                        for sub_field in n_field.sub_fields:
                            if sub_field.is_temporal:
                                continue
                            sub_field_name = to_snake_case(sub_field.name)
                            nested_row_record[sub_field_name] = nested_row.get(sub_field.name)

                        nested_row_records.append(nested_row_record)

                    sub_table_id = f"{table_name}_{field_name}"[:MAX_TABLE_NAME_LENGTH]
                    sub_rows[sub_table_id] = nested_row_records

                for nm_field in nm_relation_field_info:

                    # Only process relation if data is available in incoming row
                    if nm_field.id not in row:
                        continue

                    values = row[nm_field.id]
                    if values is not None:
                        if not isinstance(values, list):
                            values = [values]

                        field_name = to_snake_case(nm_field.name)
                        through_row_records = []
                        for value in values:
                            from_fk = id_value
                            through_row_record = {
                                f"{dataset_table.name}_id": from_fk,
                            }

                            if dataset_table.has_compound_key:
                                for id_field in dataset_table.get_fields_by_id(
                                    *dataset_table.identifier
                                ):
                                    through_row_record[
                                        f"{dataset_table.name}_{to_snake_case(id_field.name)}"
                                    ] = row[id_field.name]
                            # check is_through_table, add rows if needed
                            to_fk = value
                            if nm_field.is_through_table:
                                through_field_metas = [
                                    (f.id.split(RELATION_INDICATOR)[-1], f.is_temporal)
                                    for f in nm_field.sub_fields
                                ]
                                to_fk = ".".join(
                                    str(value[fn])
                                    for fn, is_temporal in through_field_metas
                                    if not is_temporal
                                )
                                for through_field_name, is_temporal in through_field_metas:
                                    through_field_prefix = (
                                        "" if is_temporal else f"{nm_field.name}_"
                                    )
                                    full_through_field_name = to_snake_case(
                                        f"{through_field_prefix}{through_field_name}"
                                    )
                                    through_row_record[full_through_field_name] = value[
                                        through_field_name
                                    ]
                            # PK has been changed to an autonumber
                            # through_row_record["id"] = f"{from_fk}.{to_fk}"
                            through_row_record[f"{field_name}_id"] = to_fk

                            through_row_records.append(through_row_record)

                        sub_table_id = f"{table_name}_{field_name}"[:MAX_TABLE_NAME_LENGTH]
                        sub_rows[sub_table_id] = through_row_records

                    del row[nm_field.id]
                yield {table_name: [row], **{k: cast(List[Row], v) for k, v in sub_rows.items()}}



from typing import List, Optional

from django.conf import settings
from django.core.management import BaseCommand

from schematools.contrib.django.models import Profile
from schematools.types import ProfileSchema
from schematools.utils import profile_schemas_from_url


class Command(BaseCommand):
    help = "Import all known profiles from Amsterdam schema files."
    requires_system_checks = False

    def add_arguments(self, parser) -> None:
        parser.add_argument("profile", nargs="*", help="Local profile files to import")
        parser.add_argument(
            "--schema-url",
            default=settings.PROFILES_URL,
            help=f"Schema URL (default: {settings.PROFILES_URL})",
        )

    def handle(self, *args, **options) -> None:
        if options["profile"]:
            profiles = self.import_from_files(options["profile"])
        else:
            profiles = self.import_from_url(options["schema_url"])

        if profiles:
            self.stdout.write(f"Imported profiles: {len(profiles)}")
        else:
            self.stdout.write("No new profiles imported.")

    def import_from_files(self, profile_files: List[str]) -> List[Profile]:
        profiles = []
        for filename in profile_files:
            self.stdout.write(f"Loading profile from {filename}")
            schema = ProfileSchema.from_file(filename)
            profile = self.import_schema(schema.name, schema)
            if profile is not None:
                profiles.append(profile)

        return profiles

    def import_from_url(self, schema_url: str) -> List[Profile]:
        """Import all schema definitions from an URL"""
        self.stdout.write(f"Loading profiles from {schema_url}")
        profiles = []

        for name, schema in profile_schemas_from_url(schema_url).items():
            self.stdout.write(f"* Processing {name}")
            profile = self.import_schema(name, schema)
            if profile is not None:
                profiles.append(profile)

        return profiles

    def import_schema(self, name: str, schema: ProfileSchema) -> Optional[Profile]:
        try:
            profile = Profile.objects.get(name=schema.name)
        except Profile.DoesNotExist:
            profile = Profile.create_for_schema(schema)
            self.stdout.write(f"   Created {name}")
            return profile
        else:
            updated = profile.save_for_schema(schema)
            if updated:
                self.stdout.write(f"    Updated {name}")
                return profile

        return None



"""Module to hold factories."""
from collections import defaultdict
from typing import Dict, Optional, Set

from sqlalchemy import Column, MetaData, Table

from schematools import TMP_TABLE_POSTFIX
from schematools.importer import fetch_col_type
from schematools.types import DatasetSchema
from schematools.utils import to_snake_case


def tables_factory(
    dataset: DatasetSchema,
    metadata: Optional[MetaData] = None,
    db_table_names: Optional[Dict[str, Optional[str]]] = None,
    db_schema_names: Optional[Dict[str, Optional[str]]] = None,
    limit_tables_to: Optional[Set] = None,
) -> Dict[str, Table]:
    """Generate the SQLAlchemy Table objects base on a `DatasetSchema` definition.

    Args:
        dataset: The Amsterdam Schema definition of the dataset
        metadata: SQLAlchemy schema metadata that groups all tables to a single connection.
        db_table_names: Optional sql table names, keyed on dataset_table_id.
            If not give, db_table_names are inferred from the schema name.
        db_schema_names: Optional database schema names, keyed on dataset_table_id.
            If not given, schema names default to `public`.
        limit_tables_to: Only process the indicated tables (based on table.id).

    The returned tables are keyed on the name of the dataset and table.
    SA Table objects are also created for the junction tables that are needed for relations.

    The nested and through tables that are generated on-the-fly are also taken into account.
    Special care is needed to add postfixes to tables names if the parent table of these
    tables has a non-default name. In that case, these on-the-fly tables also need to
    get an extra postfix in their db name.

    One caveat: The assumption now is that special "overridden" names for tables
    only are used because postfixes are added, and for no other reasons.
    """
    tables = defaultdict(dict)
    metadata = metadata or MetaData()

    for dataset_table in dataset.get_tables(include_nested=True, include_through=True):
        if limit_tables_to is not None and dataset_table.id not in limit_tables_to:
            continue
        db_table_description = dataset_table.description
        if (db_table_name := (db_table_names or {}).get(dataset_table.id)) is None:
            has_postfix = (
                db_schema_names is not None
                and (dataset_table.is_nested_table or dataset_table.is_through_table)
                and db_table_names.get(dataset_table.parent_table.id) is not None
            )
            postfix = TMP_TABLE_POSTFIX if has_postfix else None
            db_table_name = dataset_table.db_name(postfix=postfix)

        db_schema_name = (db_schema_names or {}).get(dataset_table.id)
        columns = []
        for field in dataset_table.fields:

            # Exclude nested and nm_relation fields (is_array check)
            # and fields that are added only for temporality
            if field.type.endswith("#/definitions/schema") or field.is_array or field.is_temporal:
                continue
            field_name = to_snake_case(field.name)
            try:
                col_type = fetch_col_type(field)
            except KeyError:
                raise NotImplementedError(
                    f'Import failed at "{field.name}": {dict(field)!r}\n'
                    f"Field type '{field.type}' is not implemented."
                ) from None
            col_kwargs = {"nullable": not field.required}
            if field.is_primary:
                col_kwargs["primary_key"] = True
                col_kwargs["nullable"] = False
                col_kwargs["autoincrement"] = field.type.endswith("autoincrement")
            id_postfix = "_id" if field.relation else ""
            columns.append(
                Column(
                    f"{field_name}{id_postfix}", col_type, comment=field.description, **col_kwargs
                )
            )

        # The table_name has to be snakecased here, because
        # it is used to lookup an SA Table and it needs to match
        # with the records that are coming back from the
        # BaseImporter.
        tables[to_snake_case(dataset_table.name)] = Table(
            db_table_name,
            metadata,
            comment=db_table_description,
            schema=db_schema_name,
            *columns,
            extend_existing=True,
        )

    return tables



from __future__ import annotations

import json
import os
import re
from functools import lru_cache
from pathlib import Path
from typing import Any, Dict, Final, Match, Optional, Pattern, Type, Union, cast

import requests
from cachetools.func import ttl_cache
from deprecated import deprecated
from more_ds.network.url import URL
from more_itertools import first
from string_utils import slugify

from schematools import MAX_TABLE_NAME_LENGTH, RELATION_INDICATOR, TMP_TABLE_POSTFIX, types

RE_CAMEL_CASE: Final[Pattern[str]] = re.compile(
    r"(((?<=[^A-Z])[A-Z])|([A-Z](?![A-Z]))|((?<=[a-z])[0-9])|(?<=[0-9])[a-z])"
)


@ttl_cache(ttl=16)
def dataset_schemas_from_url(
    schemas_url: Union[URL, str],
    dataset_name: Optional[str] = None,
    prefetch_related: bool = False,
) -> Dict[str, types.DatasetSchema]:
    """Fetch all dataset schemas from a remote file (or single dataset if specified).

    The URL could be ``https://schemas.data.amsterdam.nl/datasets/``
    """
    if dataset_name:
        schema = schema_from_url(
            base_url=schemas_url,
            data_type=types.DatasetSchema,
            dataset_id=dataset_name,
            prefetch_related=prefetch_related,
        )
        return {dataset_name: schema}

    return schemas_from_url(base_url=schemas_url, data_type=types.DatasetSchema)


def dataset_schema_from_url(
    schemas_url: Union[URL, str],
    dataset_name: str,
    prefetch_related: bool = False,
) -> types.DatasetSchema:
    """Fetch a dataset schema from a remote file."""
    return schema_from_url(
        base_url=schemas_url,
        data_type=types.DatasetSchema,
        dataset_id=dataset_name,
        prefetch_related=prefetch_related,
    )


@ttl_cache(ttl=16)
def profile_schemas_from_url(profiles_url: Union[URL, str]) -> Dict[str, types.ProfileSchema]:
    """Fetch all profile schemas from a remote file.

    The URL could be ``https://schemas.data.amsterdam.nl/profiles/``
    """
    return schemas_from_url(base_url=profiles_url, data_type=types.ProfileSchema)


def dataset_paths_from_url(base_url: Union[URL, str]) -> Dict[str, str]:
    """Fetch all dataset paths from a remote location.

    The URL could be ``https://schemas.data.amsterdam.nl/datasets/``
    """
    base_url = URL(base_url)

    with requests.Session() as connection:
        response = connection.get(base_url / "index.json")
        response.raise_for_status()
        return cast(Dict[str, str], response.json())


def schemas_from_url(base_url: Union[URL, str], data_type: Type[types.ST]) -> Dict[str, types.ST]:
    """Fetch all schema definitions from a remote file.

    The URL could be ``https://schemas.data.amsterdam.nl/datasets/``
    """
    schema_lookup: Dict[str, types.ST] = {}
    base_url = URL(base_url)

    with requests.Session() as connection:
        response = connection.get(base_url / "index.json")
        response.raise_for_status()
        response_data = response.json()

        for schema_id, schema_path in response_data.items():
            schema_lookup[schema_id] = _schema_from_url_with_connection(
                connection, base_url, schema_path, data_type
            )
    return schema_lookup


def schema_from_url(
    base_url: Union[URL, str],
    data_type: Type[types.ST],
    dataset_id: str,
    prefetch_related: bool = False,
) -> types.ST:
    """Fetch schema definitions from a remote file for a single dataset.

    The URL could be ``https://schemas.data.amsterdam.nl/datasets/``
    """
    base_url = URL(base_url)

    with requests.Session() as connection:
        index_response = connection.get(base_url / "index.json")
        index_response.raise_for_status()
        index = index_response.json()
        dataset_schema = _schema_from_url_with_connection(
            connection, base_url, index[dataset_id], data_type
        )

    # For this recursive call, we set prefetch_related=False
    # to avoid deep/endless recursion
    # The result of def_from_url does not need to be stored,
    # because is it cached on the DatasetSchema instances.
    if prefetch_related and isinstance(dataset_schema, types.DatasetSchema):
        for ds_id in dataset_schema.related_dataset_schema_ids:
            schema_from_url(base_url, data_type, ds_id, prefetch_related=False)

    return dataset_schema


def _schema_from_url_with_connection(
    connection: requests.Session, base_url: URL, dataset_path: str, data_type: Type[types.ST]
) -> types.ST:
    """Fetch single schema from url with connection."""
    response = connection.get(base_url / dataset_path / "dataset")
    response.raise_for_status()
    response_data = response.json()

    # Include referenced tables for datasets.
    for i, table in enumerate(response_data["tables"]):
        if ref := table.get("$ref"):
            table_response = connection.get(base_url / dataset_path / ref)
            table_response.raise_for_status()
            response_data["tables"][i] = table_response.json()

    schema: types.ST = data_type.from_dict(response_data)
    return schema


def dataset_schema_from_file(
    file_path: Union[Path, str], prefetch_related: bool = False
) -> types.DatasetSchema:
    """Read a dataset schema from a file on local drive.

    If `prefetch_related` is given, an index is built that maps
    the dataset id to the path towards `dataset.json` on the filesystem.
    """

    # Normalize to an absolute path.
    file_path = Path(file_path).resolve()
    index: Dict[str, str] = {}

    # Find the location of the `datasets` folder (make sure to find the "deepest" one)
    try:
        ds_root = first(
            parent for parent in reversed(file_path.parents) if parent.name == "datasets"
        )
    except ValueError:
        raise ValueError(
            f"The provided file_path ({file_path}) should contain a `datasets` folder."
        )

    # The dataset being requested
    dataset_schema = types.DatasetSchema.from_file(file_path)

    # Build the mapping from dataset -> path with jsonschema file
    if prefetch_related:
        for root, dirs, files in os.walk(ds_root):
            if "dataset.json" in set(files):
                root_path = Path(root)
                index[root_path.name] = root_path.joinpath("dataset.json")

        # Result of `schema_from_file` is discarded here, the call is only done
        # to add the dataset to the cache.
        # For this recursive call, we set prefetch_related=False
        # to avoid deep/endless recursion
        for ds_id in dataset_schema.related_dataset_schema_ids:
            dataset_schema_from_file(index[ds_id], prefetch_related=False)

    return dataset_schema


def profile_schema_from_file(filename: Union[Path, str]) -> Dict[str, types.ProfileSchema]:
    """Read a profile schema from a file on local drive."""
    with open(filename, "r") as file_handler:
        schema_info = json.load(file_handler)
        return {schema_info["name"]: types.ProfileSchema.from_dict(schema_info)}


@deprecated(
    version="1.0.4",
    reason="Does not work with datasets that have their tables split "
    "out into separate files. Use something like "
    "`schematools.cli._get_dataset_schema` instead.",
)
def schema_fetch_url_file(schema_url_file: Union[URL, str]) -> Dict[str, Any]:
    """Return schemadata from URL or File."""
    if not schema_url_file.startswith("http"):
        with open(schema_url_file) as f:
            schema_data = json.load(f)
    else:
        response = requests.get(schema_url_file)
        response.raise_for_status()
        schema_data = response.json()

    return cast(Dict[str, Any], schema_data)


_CAMEL_CASE_REPLACE_PAT: Final[Pattern[str]] = re.compile(
    r"""
    (?:_|\s)+   # Find word boundaries by looking for underscore and whitespace characters, they
                # will be discarded (not captured)
    (.)         # Capture first letter of word on word boundary
    |           # OR
    (\d+)       # Capture a number
    (?:_|\s)*   # Optionally followed by underscore and whitespace characters (to be discarded)
    (.)         # Capture first letter of word on word boundary
    """,
    re.VERBOSE,
)


@lru_cache(maxsize=500)
def toCamelCase(ident: str) -> str:
    """Convert an identifier to camelCase format.

    Word boundaries are determined by:
    - numbers
    - underscore characters
    - whitespace characters (this violates the concept of identifiers,
      but we handle it nevertheless)

    A camelCased identifier, when it starts with a letter, it will start with a lower cased letter.

    Empty strings are not allowed. They will raise an :exc:`ValueError` exception.

    Examples::

        >>> toCamelCase("dataset_table_schema")
        'datasetTableSchema'
        >>> toCamelCase("dataset table schema")
        'datasetTableSchema'
        >>> toCamelCase("fu_33_bar")
        'fu33Bar'
        >>> toCamelCase("fu_33bar")
        'fu33Bar'
        >>> toCamelCase("fu_33Bar")
        'fu33Bar'
        >>> toCamelCase("33_fu_bar")
        '33FuBar'

    Args:
        ident: The identifier to be converted.

    Returns:
        The identifier in camelCase format.

    Raises:
        ValueError: If ``indent`` is an empty string.

    """

    def replacement(m: Match) -> str:
        # As we use the OR operator in the regular expression with capture groups on both sides,
        # we will always have at least one capture group that results in `None`. We filter those
        # out in the generator expression. Even though a captured group sometimes represents a
        # number (as a string), we still call `upper()` on it. That's faster than another
        # explicit test.
        return "".join(s.upper() for s in m.groups() if s)

    if ident == "":
        raise ValueError("Parameter `ident` cannot be an empty string.")
    result = _CAMEL_CASE_REPLACE_PAT.sub(replacement, ident)
    # The first letter of camelCase identifier is always lower case
    return result[0].lower() + result[1:]


@lru_cache(maxsize=500)
def to_snake_case(ident: str) -> str:
    """Convert an identifier to snake_case format.

    Empty strings are not allowed. They will raise an :exc:`ValueError` exception.

    Args:
        ident: The identifier to be converted.

    Returns:
        The identifier in snake_case foramt.

    Raises:
        ValueError: If ``ident`` is an empty string.
    """
    if ident == "":
        raise ValueError("Parameter `ident` cannot be an empty string.")
    # Convert to field name, avoiding snake_case to snake_case issues.
    # Also preserve RELATION_INDICATOR in names (RELATION_INDICATOR are used for object relations)
    name_parts = [toCamelCase(part) for part in ident.split(RELATION_INDICATOR)]
    return RELATION_INDICATOR.join(
        slugify(RE_CAMEL_CASE.sub(r" \1", part).strip(), separator="_") for part in name_parts
    )


def get_rel_table_identifier(
    prefix_length: int, table_identifier: str, through_identifier: str
) -> str:
    """Create identifier for related table (FK or M2M) from table_identifier and extra fieldname.

    Take length of prefix (dataset.id) into account, postgresql has maxsize for tablenames.
    """
    through_table_name = f"{table_identifier}_{through_identifier}"
    return through_table_name

    # return through_table_name[: MAX_TABLE_NAME_LENGTH - len(TMP_TABLE_POSTFIX) - prefix_length]


def shorten_name(db_table_name: str, with_postfix: bool = False) -> str:
    """Shorten names to safe length for postgresql."""
    max_length = MAX_TABLE_NAME_LENGTH - int(with_postfix) * len(TMP_TABLE_POSTFIX)
    return db_table_name[:max_length]



# Generated by Django 3.0.4 on 2020-05-07 03:43

from django.db import migrations, models

import schematools.contrib.django.validators


class Migration(migrations.Migration):

    dependencies = [
        ("datasets", "0005_datasetfield"),
    ]

    operations = [
        migrations.AddField(
            model_name="dataset",
            name="enable_db",
            field=models.BooleanField(default=True),
        ),
        migrations.AddField(
            model_name="dataset",
            name="endpoint_url",
            field=models.URLField(blank=True, null=True),
        ),
        migrations.AddField(
            model_name="dataset",
            name="url_prefix",
            field=models.CharField(
                blank=True,
                max_length=100,
                validators=[schematools.contrib.django.validators.URLPathValidator()],
            ),
        ),
    ]



# Generated by Django 3.0.2 on 2020-01-29 08:35

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("datasets", "0001_initial"),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="dataset",
            options={
                "ordering": ("ordering", "name"),
                "verbose_name": "Dataset",
                "verbose_name_plural": "Datasets",
            },
        ),
        migrations.CreateModel(
            name="DatasetTable",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("name", models.CharField(max_length=100)),
                ("enable_geosearch", models.BooleanField(default=True)),
                ("db_table", models.CharField(max_length=100, unique=True)),
                (
                    "display_field",
                    models.CharField(blank=True, max_length=50, null=True),
                ),
                (
                    "geometry_field",
                    models.CharField(blank=True, max_length=50, null=True),
                ),
                (
                    "dataset",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="tables",
                        to="datasets.Dataset",
                    ),
                ),
            ],
            options={
                "verbose_name": "Dataset Table",
                "verbose_name_plural": "Dataset Tables",
                "ordering": ("name",),
                "unique_together": {("dataset", "name")},
            },
        ),
    ]



from schematools.types import DatasetSchema

from .generators.mapfile import MapfileGenerator

# XXX see generators/mapfile.py about this import
from .interfaces.mapfile.serializers import MappyfileSerializer


def create_map_from_dataset(dataset: DatasetSchema):
    """Creates a Mapfile from a dataset in JSON"""

    _generator = MapfileGenerator(serializer=MappyfileSerializer())
    return _generator(dataset)



"""Module implementing an event processor, that processes full events."""
from __future__ import annotations

import json
import logging
from collections import defaultdict
from typing import Dict, List

from sqlalchemy.engine import Connection

from schematools.events import metadata
from schematools.factories import tables_factory
from schematools.types import DatasetSchema
from schematools.utils import to_snake_case, toCamelCase

# Enable the sqlalchemy logger by uncommenting the following 2 lines to debug SQL related issues
# logging.basicConfig()
# logging.getLogger("sqlalchemy.engine").setLevel(logging.INFO)

logger = logging.getLogger(__name__)


# Configuration information to map the event-type
# to the following fields:
# - db_operation_name (INSERT/UPDATE)
# - needs_select: generated SQL needs to select a record

EVENT_TYPE_MAPPINGS = {
    "ADD": ("insert", False),
    "MODIFY": ("update", True),
    "DELETE": ("delete", True),
}


class DataSplitter:
    """Helper class to split a data event record.

    Data is split into scalar fields and relational fields.
    Scalar data can be used directly to provide SQL queries with data.
    Relational data is used to update records with FK colums, or, for NM relations,
    to construct SQL statement to update junction tables.
    """

    def __init__(
        self, events_processor: EventsProcessor, dataset_id: str, table_id: str, event_data: dict
    ) -> None:
        """Construct the DataSplitter.

        Args:
            events_processor: reference to the EventsProcessor, usually,
                this is a backref. where the EventsProcessor is instantiating the DataSplitter
            dataset_id: identifier of the dataset
            table_id: identifier of the table
            event_data: the actual event data
        """
        self.events_processor = events_processor
        self.dataset_id = dataset_id
        self.table_id = table_id
        self.dataset_table = events_processor.datasets[dataset_id].get_table_by_id(table_id)
        self.junction_table_rows = {}
        relational_data, self.row_data = self._split_data(event_data)
        self.process_relational_data(relational_data)

    def _split_data(self, event_data: dict) -> list[dict]:
        """Split the event_data in 2 parts, one with scalars, the other with relations."""
        data_bags = [{}, {}]
        for field in self.dataset_table.fields:
            snaked_field_id = to_snake_case(field.id)
            if snaked_field_id in event_data:
                data_bags[field.is_scalar][snaked_field_id] = event_data[snaked_field_id]
        return data_bags

    def process_relational_data(self, relational_data: dict) -> None:
        """Process the relational_data.

        Determines if relation is FK or NM. Handles appropriately.
        """
        for field_id, field_data in relational_data.items():
            field = self.dataset_table.get_field_by_id(toCamelCase(field_id))
            if field.relation is not None:
                self._handle_fk(field_id, field_data, field.relation)
                field_data = [] if field_data is None else [field_data]
                self._handle_junction_table(field_id, field_data, field.relation)
            elif field.nm_relation is not None:
                self._handle_junction_table(field_id, field_data, field.nm_relation)
            else:
                raise Exception("Relation should be either FK or NM")

    def fetch_row(self) -> dict:
        """Fetch the row data.

        To be used in a SQL update or insert query.
        """
        return self.row_data

    def _handle_fk(self, field_id: str, field_data: dict, relation: str):
        """Handle FK relation."""
        fk_row_data = {}
        snaked_field_id = to_snake_case(field_id)
        target_identifier_fields = self._fetch_target_identifier_fields(relation)

        # id for target side of relation
        if field_data:
            id_value = ".".join((str(field_data[fn]) for fn in target_identifier_fields))
        else:
            id_value = None

        fk_row_data[f"{snaked_field_id}_id"] = id_value
        for fn in self._fetch_target_identifier_fields(relation):
            fk_row_data[f"{snaked_field_id}_{fn}"] = (field_data or {}).get(fn)

        self.row_data.update(fk_row_data)

    def _fetch_target_identifier_fields(self, relation: str) -> list[str]:
        """Fetch the identifier fields for the target side of a relation.

        These fields will be ordered with the main identifier first, and
        on the second position the sequence identifier.
        """
        dataset_id, table_id = relation.split(":")
        target_dataset = self.events_processor.datasets[dataset_id]
        target_dataset_table = target_dataset.get_table_by_id(table_id)
        return target_dataset_table.identifier

    def _fetch_source_id_info(self):
        # id info for source side of relation (for updates/deletes)
        snaked_source_table = to_snake_case(self.table_id)
        identifier = self.dataset_table.identifier
        id_part_values = {fn: self.row_data[fn] for fn in identifier}
        id_value = ".".join(str(part) for part in id_part_values.values())
        return (f"{snaked_source_table}_id", id_value, id_part_values)

    def _handle_junction_table(self, field_id: str, field_data: dict, nm_relation):
        """Handle NM relation."""
        snaked_field_id = to_snake_case(field_id)
        target_identifier_fields = self._fetch_target_identifier_fields(nm_relation)
        snaked_source_table_id, id_value, id_part_values = self._fetch_source_id_info()

        # Short circuit when no data
        if not field_data:
            self.junction_table_rows[snaked_field_id] = []
            return

        nm_rows = []

        for row_data in field_data:
            nm_row_data = {}

            # relation fields for source side
            # XXX add support for shortnames!
            for sub_field_schema in self.dataset_table.get_fields_by_id(
                *self.dataset_table.identifier
            ):
                sub_field_id = f"{self.table_id}_{sub_field_schema.id}"
                nm_row_data[sub_field_id] = id_part_values[sub_field_schema.id]

            # relation fields for target side
            for fn, fv in row_data.items():
                name_prefix = f"{snaked_field_id}_" if fn in target_identifier_fields else ""
                nm_row_data[f"{name_prefix}{fn}"] = fv

            # id for source side of relation
            nm_row_data[snaked_source_table_id] = id_value

            # id for target side of relation
            target_id_value = ".".join((str(row_data[fn]) for fn in target_identifier_fields))
            nm_row_data[f"{snaked_field_id}_id"] = target_id_value

            # PK field, needed by Django
            # PK has been changed to an autonumber
            # nm_row_data["id"] = f"{id_value}.{target_id_value}"

            nm_rows.append(nm_row_data)

        self.junction_table_rows[snaked_field_id] = nm_rows

    def update_relations(self):
        """Perform SQL calls to update the relations.

        Update is performed by first deleting all records pointing
        to the source table, and then re-inserting those records.

        Assertion is that GOB always provides fully populated records.
        """
        conn = self.events_processor.conn

        with conn.begin():
            for snaked_field_id, nm_row_data in self.junction_table_rows.items():

                snaked_source_table_id = to_snake_case(self.table_id)
                junction_table_id = f"{snaked_source_table_id}_{snaked_field_id}"
                sa_table = self.events_processor.tables[self.dataset_id][junction_table_id]
                source_id, source_id_value, _ = self._fetch_source_id_info()

                source_id_column = getattr(sa_table.c, source_id)
                conn.execute(sa_table.delete().where(source_id_column == source_id_value))
                if nm_row_data:
                    conn.execute(sa_table.insert(), nm_row_data)


class EventsProcessor:
    """The core event processing class.

    It needs to be initialised once
    with configuration (datasets) and a db connection.
    Once initialised, the process_event() method is able to
    process incoming events.
    The database actions are done using SQLAlchemy Core. So,
    a helper function `tables_factory()` is used to created the
    SA Tables that are needed during the processing of the events.
    """

    def __init__(
        self,
        datasets: List[DatasetSchema],
        srid: str,
        connection: Connection,
        local_metadata=None,
        truncate=False,
    ) -> None:
        """Construct the event processor.

        Args:
            datasets: list of DatasetSchema instances. Usually dataset tables
                have relations to tables.
                If these target tables are in different datasets,
                these dataset also need to be provided.
            srid: coordinate system
            local_metadata: SQLAlchemy metadata object, only needs to be provided
                in unit tests.
            truncate: indicates if the relational tables need to be truncated
        """
        self.datasets: Dict[str, DatasetSchema] = {ds.id: ds for ds in datasets}
        self.srid = srid
        self.conn = connection
        _metadata = local_metadata or metadata  # mainly for testing
        _metadata.bind = connection.engine
        self.tables = {}
        for dataset_id, dataset in self.datasets.items():
            base_tables_ids = {dataset_table.id for dataset_table in dataset.tables}
            self.tables[dataset_id] = tfac = tables_factory(dataset, metadata=_metadata)
            self.geo_fields = defaultdict(lambda: defaultdict(list))
            for table_id, table in tfac.items():
                if not table.exists():
                    table.create()
                elif truncate:
                    with self.conn.begin():
                        self.conn.execute(table.delete())
                # self.has_compound_key = dataset_table.has_compound_key
                # skip the generated nm tables
                if table_id not in base_tables_ids:
                    continue
                for field in dataset.get_table_by_id(table_id).fields:
                    if field.is_geo:
                        self.geo_fields[dataset_id][table_id].append(field.name)

    def process_row(self, event_id: str, event_meta: dict, event_data: dict) -> None:
        """Process one row of data.

        Args:
            event_id: Id of the event (Kafka id)
            event_meta: Metadata about the event
            event_data: Data containing the fields of the event
        """
        event_type = event_meta["event_type"]
        db_operation_name, needs_select = EVENT_TYPE_MAPPINGS[event_type]
        dataset_id = event_meta["dataset_id"]
        table_id = event_meta["table_id"]
        data_splitter = DataSplitter(self, dataset_id, table_id, event_data)

        row = data_splitter.fetch_row()

        for field_name in self.geo_fields[dataset_id][table_id]:
            geo_value = row.get(field_name)
            if geo_value is not None and not row[field_name].startswith("SRID"):
                row[field_name] = f"SRID={self.srid};{geo_value}"

        identifier = self.datasets[dataset_id].get_table_by_id(table_id).identifier
        id_value = ".".join(str(row[fn]) for fn in identifier)
        row["id"] = id_value

        table = self.tables[dataset_id][table_id]
        db_operation = getattr(table, db_operation_name)()
        if needs_select:
            # XXX Can we assume 'id' is always available?
            db_operation = db_operation.where(table.c.id == id_value)
        with self.conn.begin():
            self.conn.execute(db_operation, row)

        # now process the relations (if any)
        data_splitter.update_relations()

    def process_event(self, event_id: str, event_meta: dict, event_data: dict):
        """Do inserts/updates/deletes."""
        self.process_row(event_id, event_meta, event_data)

    def load_events_from_file(self, events_path: str):
        """Load events from a file, primarily used for testing."""
        with open(events_path) as ef:
            for line in ef:
                if line.strip():
                    event_id, event_meta_str, data_str = line.split("|", maxsplit=2)
                    event_meta = json.loads(event_meta_str)
                    event_data = json.loads(data_str)
                    self.process_event(
                        event_id,
                        event_meta,
                        event_data,
                    )



import numbers
from decimal import Decimal

from geoalchemy2 import Geometry
from sqlalchemy import BigInteger, Boolean, Date, DateTime, Float, Numeric, String, Time
from sqlalchemy.types import ARRAY

from schematools.types import DatasetTableSchema

FORMAT_MODELS_LOOKUP = {
    "date": Date,
    "time": Time,
    "date-time": DateTime,
    "uri": String,
    "email": String,
}

JSON_TYPE_TO_PG = {
    "string": String,
    "object": String,
    "boolean": Boolean,
    "integer": BigInteger,
    "integer/autoincrement": BigInteger,
    "number": Float,
    "array": ARRAY(String),
    "https://schemas.data.amsterdam.nl/schema@v1.1.0#/definitions/id": String,
    "https://schemas.data.amsterdam.nl/schema@v1.1.0#/definitions/class": String,
    "https://schemas.data.amsterdam.nl/schema@v1.1.0#/definitions/dataset": String,
    "https://schemas.data.amsterdam.nl/schema@v1.1.0#/definitions/schema": String,
    "https://schemas.data.amsterdam.nl/schema@v1.1.1#/definitions/id": String,
    "https://schemas.data.amsterdam.nl/schema@v1.1.1#/definitions/class": String,
    "https://schemas.data.amsterdam.nl/schema@v1.1.1#/definitions/dataset": String,
    "https://schemas.data.amsterdam.nl/schema@v1.1.1#/definitions/schema": String,
    "https://geojson.org/schema/Geometry.json": Geometry(geometry_type="GEOMETRY", srid=28992),
    "https://geojson.org/schema/Point.json": Geometry(geometry_type="POINT", srid=28992),
    "https://geojson.org/schema/Polygon.json": Geometry(geometry_type="POLYGON", srid=28992),
    "https://geojson.org/schema/MultiPolygon.json": Geometry(
        geometry_type="MULTIPOLYGON", srid=28992
    ),
    "https://geojson.org/schema/MultiPoint.json": Geometry(geometry_type="MULTIPOINT", srid=28992),
    "https://geojson.org/schema/LineString.json": Geometry(geometry_type="LINESTRING", srid=28992),
    "https://geojson.org/schema/MultiLineString.json": Geometry(
        geometry_type="MULTILINESTRING", srid=28992
    ),
}


def numeric_datatype_scale(scale_=None):
    """detect scale from decimal for database datatype scale definition"""
    if (
        isinstance(scale_, numbers.Number)
        and str(scale_).count("1") == 1
        and str(scale_).endswith("1")
    ):
        # TODO: make it possible to set percision too
        # now it defaults to max of 12
        get_scale = Decimal(str(scale_)).as_tuple().exponent
        if get_scale < 0:
            get_scale = get_scale * -1
        return Numeric(precision=12, scale=get_scale)
    else:
        return Numeric


def fetch_col_type(field):
    col_type = JSON_TYPE_TO_PG[field.type]
    if (field_format := field.format) is not None:
        return FORMAT_MODELS_LOOKUP[field_format]
    # TODO: format takes precedence over multipleof
    # if there is an use case that both can apply for a field definition
    # then logic must be changed
    field_multiple = field.multipleof
    if field_multiple is not None:
        return numeric_datatype_scale(scale_=field_multiple)
    return col_type


def get_table_name(dataset_table: DatasetTableSchema) -> str:
    """Generate the database identifier for the table."""
    schema = dataset_table._parent_schema
    return f"{schema.id}_{dataset_table.id}".replace("-", "_")



"""Cli tools."""

import sys
from collections import defaultdict
from typing import Any, DefaultDict, Dict, Iterable, List, Optional, Tuple

import click
import jsonschema
import requests
import sqlalchemy
from deepdiff import DeepDiff
from json_encoder import json
from sqlalchemy import create_engine
from sqlalchemy.exc import SQLAlchemyError

from schematools.db import (
    create_meta_table_data,
    create_meta_tables,
    fetch_schema_from_relational_schema,
    fetch_table_names,
)
from schematools.events.export import export_events
from schematools.events.full import EventsProcessor
from schematools.exceptions import ParserError
from schematools.importer.base import BaseImporter
from schematools.importer.geojson import GeoJSONImporter
from schematools.importer.ndjson import NDJSONImporter
from schematools.introspect.db import introspect_db_schema
from schematools.introspect.geojson import introspect_geojson_files
from schematools.maps import create_mapfile
from schematools.permissions.db import (
    apply_schema_and_profile_permissions,
    introspect_permissions,
    revoke_permissions,
)
from schematools.provenance.create import ProvenanceIteration
from schematools.types import DatasetSchema, SchemaType
from schematools.utils import (
    dataset_schema_from_file,
    dataset_schema_from_url,
    dataset_schemas_from_url,
    schema_fetch_url_file,
)
from schematools.validation import Validator

DEFAULT_SCHEMA_URL = "https://schemas.data.amsterdam.nl/datasets/"
DEFAULT_PROFILE_URL = "https://schemas.data.amsterdam.nl/profiles/"

option_db_url = click.option(
    "--db-url",
    envvar="DATABASE_URL",
    required=True,
    help="DSN of database, can also use DATABASE_URL environment.",
)
option_schema_url = click.option(
    "--schema-url",
    envvar="SCHEMA_URL",
    default=DEFAULT_SCHEMA_URL,
    show_default=True,
    required=True,
    help="Url where valid amsterdam schema files are found. "
    "SCHEMA_URL can also be provided as environment variable.",
)

argument_schema_location = click.argument(
    "schema_location",
    metavar="(DATASET-ID | DATASET-FILENAME)",
)

option_profile_url = click.option(
    "--profile-url",
    envvar="PROFILE_URL",
    default=DEFAULT_PROFILE_URL,
    show_default=True,
    required=True,
    help="Url where valid amsterdam profile files are found. "
    "PROFILE_URL can also be provided as environment variable.",
)

argument_profile_location = click.argument(
    "profile_location",
    metavar="(PROFILE-FILENAME | NONE)",
)

argument_role = click.argument(
    "role",
)


def _get_engine(db_url: str, pg_schemas: Optional[List[str]] = None) -> sqlalchemy.engine.Engine:
    """Initialize the SQLAlchemy engine, and report click errors."""
    kwargs = {}
    if pg_schemas is not None:
        csearch_path = ",".join(pg_schemas + ["public"])
        kwargs["connect_args"] = {"options": f"-csearch_path={csearch_path}"}
    try:
        return create_engine(db_url, **kwargs)
    except SQLAlchemyError as e:
        raise click.BadParameter(str(e), param_hint="--db-url") from e


def main() -> None:
    """Main entry point.

    This catches relevant errors, so the user is not
    confronted with internal tracebacks.
    """
    try:
        schema()
    except (EnvironmentError, SQLAlchemyError, ParserError) as e:
        click.echo(f"{e.__class__.__name__}: {e}", err=True)
        exit(1)


@click.group()
def schema() -> None:
    """Command line utility to work with Amsterdam Schema files."""
    pass


@schema.group("import")
def import_() -> None:
    """Subcommand to import data."""
    pass


@schema.group("export")
def export() -> None:
    """Subcommand to export data."""
    pass


@schema.group()
def show() -> None:
    """Show existing metadata"""
    pass


@schema.group()
def permissions() -> None:
    """Show existing metadata"""
    pass


@schema.group()
def kafka() -> None:
    """Subcommand to consume or produce kafka events."""
    pass


@permissions.command("introspect")
@option_db_url
@argument_role
def permissions_introspect(db_url: str, role: str) -> None:
    """Retrieve ACLs from a database."""
    engine = _get_engine(db_url)
    introspect_permissions(engine, role)


@permissions.command("revoke")
@option_db_url
@argument_role
def permissions_revoke(db_url: str, role: str) -> None:
    """Revoke all table select priviliges for role."""
    engine = _get_engine(db_url)
    revoke_permissions(engine, role)


@permissions.command("apply")
@option_db_url
@option_schema_url
@option_profile_url
@click.option(
    "--schema-filename",
    is_flag=False,
    help="Filename of local Amsterdam Schema (single dataset)."
    " If specified, it will be used instead of schema-url",
)
@click.option(
    "--profile-filename",
    is_flag=False,
    help="Filename of local Profile. If specified, it will be used instead of profile-url",
)
@click.option(
    "--pg_schema",
    is_flag=False,
    default="public",
    show_default=True,
    help="Postgres schema containing the data",
)
@click.option(
    "--auto",
    is_flag=True,
    default=False,
    help="Grant each scope X to their associated db role scope_x.",
)
@click.option(
    "--role",
    is_flag=False,
    default="",
    help="Role to receive grants. Ignored when --auto=True",
)
@click.option(
    "--scope",
    is_flag=False,
    default="",
    help="Scope to be granted. Ignored when --auto=True",
)
@click.option(
    "--execute/--dry-run",
    default=False,
    help="Execute SQL statements or dry-run [default]",
)
@click.option(
    "--read/--no-read",
    "set_read_permissions",
    default=True,
    help="Set read permissions [default=read]",
)
@click.option(
    "--write/--no-write",
    "set_write_permissions",
    default=True,
    help="Set dataset-level write permissions [default=write]",
)
@click.option("--create-roles", is_flag=True, default=False, help="Create missing postgres roles")
@click.option(
    "--revoke",
    is_flag=True,
    default=False,
    help="Before granting new permissions, revoke first all previous table and column permissions",
)
def permissions_apply(
    db_url: str,
    schema_url: str,
    profile_url: str,
    schema_filename: str,
    profile_filename: str,
    pg_schema: str,
    auto: bool,
    role: str,
    scope: str,
    execute: bool,
    create_roles: bool,
    revoke: bool,
    set_read_permissions: bool,
    set_write_permissions: bool,
) -> None:
    """Set permissions for a postgres role,
    based on a scope from Amsterdam Schema or Profiles."""
    dry_run = not execute

    if auto:
        role = "AUTO"
        scope = "ALL"

    engine = _get_engine(db_url)

    if schema_filename:
        dataset_schema = DatasetSchema.from_file(schema_filename)
        ams_schema = {dataset_schema.id: dataset_schema}
    else:
        ams_schema = dataset_schemas_from_url(schemas_url=schema_url)

    if profile_filename:
        profile = schema_fetch_url_file(profile_filename)
        profiles = {profile["name"]: profile}
    else:
        # Profiles not live yet, temporarilly commented out
        # profiles = profile_defs_from_url(profiles_url=profile_url)
        profiles = None

    if auto or (role and scope):
        apply_schema_and_profile_permissions(
            engine,
            pg_schema,
            ams_schema,
            profiles,
            role,
            scope,
            set_read_permissions,
            set_write_permissions,
            dry_run,
            create_roles,
            revoke,
        )
    else:
        print(
            "Choose --auto or specify both a --role and a --scope to be able to grant permissions"
        )


@schema.group()
def introspect() -> None:
    """Subcommand to generate a schema."""
    pass


@schema.group()
def create() -> None:
    """Subcommand to create a DB object."""
    pass


@schema.group()
def diff() -> None:
    """Subcommand to show schema diffs."""
    pass


def _fetch_json(location: str) -> Dict[str, Any]:
    """Fetch JSON from file or URL.

    Args:
        location: a file name or an URL

    Returns:
        JSON data as a dictionary.
    """
    if not location.startswith("http"):
        with open(location) as f:
            json_obj = json.load(f)
    else:
        response = requests.get(location)
        response.raise_for_status()
        json_obj = response.json()
    return json_obj


@schema.command()
@option_schema_url
@argument_schema_location
@click.option(
    "--additional-schemas",
    "-a",
    multiple=True,
    help="Id of a dataset schema that will be preloaded. "
    "To be used mainly for schemas that are related to the schema that is being validated.",
)
@click.argument("meta_schema_url")
def validate(
    schema_url: str, schema_location: str, additional_schemas: List[str], meta_schema_url: str
) -> None:
    """Validate a JSON file against the amsterdam schema meta schema.
    schema_location can be a url or a filesystem path.

    DATASET-ID | DATASET-FILENAME: When an DATASET-ID is provided, this is combined with
    the schema-url to produce a full url to a dataset schema. A DATASET-FILENAME is detected
    when the argument has a '.' or a '/'.

    META_SCHEMA_URL is the url where the metaschema for amsterdam schema definitions can be found.
    """

    meta_schema = _fetch_json(meta_schema_url)
    dataset = _get_dataset_schema(schema_location, schema_url, prefetch_related=True)

    # The additional schemas are fetched, but the result is not used
    # because the only reason to fetch the additional schemas is to have those schemas
    # available in the cache that is part of the DatasetSchema class
    for schema in additional_schemas:
        _get_dataset_schema(schema, schema_url)

    structural_errors = False
    try:
        click.echo("Structural validation: ", nl=False)
        jsonschema.validate(instance=dataset.json_data(), schema=meta_schema)
    except (jsonschema.ValidationError, jsonschema.SchemaError) as e:
        structural_errors = True
        click.echo(f"\n{e!s}", err=True)
    else:
        click.echo("success!")

    click.echo("Semantic validation: ", nl=False)
    semantic_errors = False
    validator = Validator(dataset=dataset)
    for error in validator.run_all():
        semantic_errors = True
        click.echo(f"\n{error!s}", err=True)
    if not semantic_errors:
        click.echo("success!")

    if structural_errors or semantic_errors:
        sys.exit(1)


@schema.command()
@click.argument("meta_schema_url")
@click.argument("schema_files", nargs=-1)
def batch_validate(meta_schema_url: str, schema_files: Tuple[str]) -> None:
    """Batch validate schema's.

    This command was tailored so that it could be run from a pre-commit hook. As a result,
    the order and type of its arguments differ from other `schema` sub-commands.

    It will perform both structural and semantic validation of the schema's.

    \b
    META_SCHEMA_URL: the URL to the Amsterdam meta schema
    SCHEMA_FILES: one or more schema files to be validated
    """
    meta_schema = _fetch_json(meta_schema_url)
    errors: DefaultDict[str, List[str]] = defaultdict(list)
    for schema in schema_files:
        try:
            dataset = _get_dataset_schema(schema, prefetch_related=True)
        except ValueError as ve:
            errors[schema].append(str(ve))
            # No sense in continuing if we can't read the schema file.
            break
        try:
            jsonschema.validate(instance=dataset.json_data(), schema=meta_schema)
        except (jsonschema.ValidationError, jsonschema.SchemaError) as struct_error:
            errors[schema].append(f"{struct_error.message}: ({', '.join(struct_error.path)})")

        validator = Validator(dataset=dataset)
        for sem_error in validator.run_all():
            errors[schema].append(str(sem_error))
    if errors:
        width = len(max(errors.keys()))
        for schema, error_messages in errors.items():
            for err_msg in error_messages:
                print(f"{schema:>{width}}: {err_msg}")
        sys.exit(1)


@show.command("provenance")
@click.argument("schema_location")
def show_provenance(schema_location: str) -> None:
    """Retrieve the key-values pairs of the source column
    (specified as a 'provenance' property of an attribute)
    and its translated name (the attribute name itself)"""
    dataset = _get_dataset_schema(schema_location, prefetch_related=True)
    try:
        instance = ProvenanceIteration(dataset)
        click.echo(instance.final_dic)
    except (jsonschema.ValidationError, jsonschema.SchemaError, KeyError) as e:
        click.echo(str(e), err=True)
        exit(1)


@show.command("tablenames")
@option_db_url
def show_tablenames(db_url: str) -> None:
    """Retrieve tablenames from a database."""
    engine = _get_engine(db_url)
    click.echo("\n".join(fetch_table_names(engine)))


@show.command("schema")
@option_db_url
@click.argument("dataset_id")
def show_schema(db_url: str, dataset_id: str) -> None:
    """Generate a json schema based on a schema define in a relational database."""
    engine = _get_engine(db_url, pg_schemas=["meta"])
    aschema = fetch_schema_from_relational_schema(engine, dataset_id)
    click.echo(json.dumps(aschema, indent=2))


@show.command("mapfile")
@click.option(
    "--schema-url",
    envvar="SCHEMA_URL",
    default=DEFAULT_SCHEMA_URL,
    show_default=True,
    required=True,
    help="Url where valid amsterdam schema files are found. "
    "SCHEMA_URL can also be provided as environment variable.",
)
@click.argument("dataset_id")
def show_mapfile(schema_url: str, dataset_id: str) -> None:
    """Generate a mapfile based on a dataset schema."""
    try:
        dataset_schema = dataset_schema_from_url(schema_url, dataset_id)
    except KeyError:
        raise click.BadParameter(f"Schema {dataset_id} not found.") from None
    click.echo(create_mapfile(dataset_schema))


@introspect.command("db")
@click.option("--db-schema", "-s", help="Tables are in a different postgres schema (not 'public')")
@click.option("--prefix", "-p", help="Tables have prefix that needs to be stripped")
@option_db_url
@click.argument("dataset_id")
@click.argument("tables", nargs=-1)
def introspect_db(
    db_schema: str, prefix: str, db_url: str, dataset_id: str, tables: Iterable[str]
) -> None:
    """Generate a schema for the tables in a database"""
    engine = _get_engine(db_url)
    aschema = introspect_db_schema(engine, dataset_id, tables, db_schema, prefix)
    click.echo(json.dumps(aschema, indent=2))


@introspect.command("geojson")
@click.argument("dataset_id")
@click.argument("files", nargs=-1, required=True)
def introspect_geojson(dataset_id: str, files: Iterable[str]) -> None:
    """Generate a schema from a GeoJSON file."""
    aschema = introspect_geojson_files(dataset_id, files)
    click.echo(json.dumps(aschema, indent=2))


@import_.command("ndjson")
@option_db_url
@option_schema_url
@argument_schema_location
@click.argument("table_name")
@click.argument("ndjson_path")
@click.option("--truncate-table", is_flag=True)
def import_ndjson(
    db_url: str,
    schema_url: str,
    schema_location: str,
    table_name: str,
    ndjson_path: str,
    truncate_table: bool,
) -> None:
    """Import a NDJSON file into a table."""
    engine = _get_engine(db_url)
    dataset_schema = _get_dataset_schema(schema_location, schema_url)
    importer = NDJSONImporter(dataset_schema, engine)
    importer.load_file(ndjson_path, table_name, truncate=truncate_table)


@import_.command("geojson")
@option_db_url
@option_schema_url
@argument_schema_location
@click.argument("table_name")
@click.argument("geojson_path")
@click.option("--truncate-table", is_flag=True)
def import_geojson(
    db_url: str,
    schema_url: str,
    schema_location: str,
    table_name: str,
    geojson_path: str,
    truncate_table: bool,
) -> None:
    """Import a GeoJSON file into a table."""
    engine = _get_engine(db_url)
    dataset_schema = _get_dataset_schema(schema_location, schema_url)
    importer = GeoJSONImporter(dataset_schema, engine)
    importer.load_file(geojson_path, table_name, truncate=truncate_table)


@import_.command("events")
@option_db_url
@option_schema_url
@argument_schema_location
@click.option("--additional-schemas", "-a", multiple=True)
@click.argument("events_path")
@click.option("-t", "--truncate-table", default=False, is_flag=True)
def import_events(
    db_url: str,
    schema_url: str,
    schema_location: str,
    additional_schemas: str,
    events_path: str,
    truncate_table: bool,
) -> None:
    """Import an events file into a table."""
    engine = _get_engine(db_url)
    dataset_schemas = [_get_dataset_schema(schema_location, schema_url)]
    for schema in additional_schemas:
        dataset_schemas.append(_get_dataset_schema(schema, schema_url))
    srid = dataset_schemas[0]["crs"].split(":")[-1]
    # Create connection, do not start a transaction.
    with engine.connect() as connection:
        importer = EventsProcessor(dataset_schemas, srid, connection, truncate=truncate_table)
        importer.load_events_from_file(events_path)


@import_.command("schema")
@option_db_url
@option_schema_url
@argument_schema_location
def import_schema(db_url: str, schema_url: str, schema_location: str) -> None:
    """Import the schema definition into the local database."""
    # Add drop or not flag
    engine = _get_engine(db_url)
    dataset_schema = _get_dataset_schema(schema_location, schema_url)

    create_meta_tables(engine)
    create_meta_table_data(engine, dataset_schema)


def _get_dataset_schema(
    schema_location: str, schema_url: Optional[str] = None, prefetch_related: bool = False
) -> DatasetSchema:
    """Find the dataset schema for the given dataset"""
    if "." in schema_location or "/" in schema_location:
        click.echo(f"Reading schema from {schema_location}", err=True)
        return dataset_schema_from_file(schema_location, prefetch_related=prefetch_related)
    else:
        assert schema_url is not None, "We should use a more robust filename detection."
        # Read the schema from the online repository.
        click.echo(f"Reading schemas from {schema_url}", err=True)

        try:
            return dataset_schema_from_url(
                schema_url, schema_location, prefetch_related=prefetch_related
            )
        except KeyError:
            raise click.BadParameter(f"Schema {schema_location} not found.") from None


@create.command("extra_index")
@option_db_url
@option_schema_url
@argument_schema_location
def create_identifier_index(db_url: str, schema_url: str, schema_location: str) -> None:
    """Execute SQLalchemy Index based on Identifier in the JSON schema data defintion"""
    engine = _get_engine(db_url)
    dataset_schema = _get_dataset_schema(schema_location, schema_url)
    importer = BaseImporter(dataset_schema, engine)

    for table in dataset_schema.tables:
        importer.generate_db_objects(table.id, ind_tables=False, ind_extra_index=True)


@create.command("tables")
@option_db_url
@option_schema_url
@argument_schema_location
def create_tables(db_url: str, schema_url: str, schema_location: str) -> None:
    """Execute SQLalchemy Table objects"""
    engine = _get_engine(db_url)
    dataset_schema = _get_dataset_schema(schema_location, schema_url, prefetch_related=True)
    importer = BaseImporter(dataset_schema, engine)

    for table in dataset_schema.tables:
        importer.generate_db_objects(table.id, ind_extra_index=False, ind_tables=True)


@create.command("all")
@option_db_url
@option_schema_url
@argument_schema_location
def create_all_objects(db_url: str, schema_url: str, schema_location: str) -> None:
    """Execute SQLalchemy Index (Identifier fields) and Table objects."""
    engine = _get_engine(db_url)
    dataset_schema = _get_dataset_schema(schema_location, schema_url, prefetch_related=True)
    importer = BaseImporter(dataset_schema, engine)

    for table in dataset_schema.tables:
        importer.generate_db_objects(table.id)


@diff.command("all")
@option_schema_url
@click.argument("diff_schema_url")
def diff_schemas(schema_url: str, diff_schema_url: str) -> None:
    """Show diff for two sets of schemas.

    The left-side schemas location is
    defined in SCHEMA_URL (or via --schema-url), the right-side schemas location
    has to be on the command-line.

    This can be used to compare two sets of schemas, e.g. ACC and PRD schemas.

    For nicer output, pipe it through a json formatter.
    """
    schemas = dataset_schemas_from_url(schema_url)
    diff_schemas = dataset_schemas_from_url(diff_schema_url)
    click.echo(DeepDiff(schemas, diff_schemas, ignore_order=True).to_json())


@export.command("events")
@option_db_url
@option_schema_url
@argument_schema_location
@click.option("--additional-schemas", "-a", multiple=True)
@click.argument("dataset_id")
@click.argument("table_id")
def export_events_for(
    db_url: str,
    schema_url: str,
    schema_location: str,
    additional_schemas: str,
    dataset_id: str,
    table_id: str,
) -> None:
    """Export events from postgres."""
    engine = _get_engine(db_url)
    dataset_schemas = [_get_dataset_schema(schema_location, schema_url)]
    for schema in additional_schemas:
        dataset_schemas.append(_get_dataset_schema(schema, schema_url))
    # Run as a transaction
    with engine.begin() as connection:
        for event in export_events(dataset_schemas, dataset_id, table_id, connection):
            click.echo(event)


@kafka.command()
@option_db_url
@option_schema_url
@argument_schema_location
@click.option("--additional-schemas", "-a", multiple=True)
@click.option("--topics", "-t", multiple=True)
@click.option("--truncate-table", default=False, is_flag=True)
def consume(
    db_url: str,
    schema_url: str,
    schema_location: str,
    additional_schemas: str,
    topics: Iterable[str],
    truncate_table: bool,
) -> None:
    """Consume kafka events."""
    # Late import, to prevent dependency on confluent-kafka for every cli user
    from schematools.events.consumer import consume_events

    engine = _get_engine(db_url)
    dataset_schemas = [_get_dataset_schema(schema_location, schema_url)]
    for schema in additional_schemas:
        dataset_schemas.append(_get_dataset_schema(schema, schema_url))
    srid = dataset_schemas[0]["crs"].split(":")[-1]
    # Create connection, do not start a transaction.
    with engine.connect() as connection:
        consume_events(dataset_schemas, srid, connection, topics, truncate=truncate_table)


if __name__ == "__main__":
    main()



"""Datbase storage of metadata from imported Amsterdam schema files."""
import json
from collections import defaultdict

from dateutil.parser import parse as dtparse
from sqlalchemy import DateTime, inspect
from sqlalchemy.orm import sessionmaker
from string_utils import camel_case_to_snake, snake_case_to_camel

from schematools import models
from schematools.exceptions import ParserError
from schematools.types import DatasetSchema
from schematools.utils import toCamelCase


def fetch_table_names(engine):
    """Fetches all tablenames, to be used in other commands"""
    insp = inspect(engine)
    return insp.get_table_names()


def create_meta_tables(engine):
    models.Base.metadata.drop_all(engine)
    models.Base.metadata.create_all(engine)


def transformer_factory(model):

    transforms = defaultdict(lambda: lambda x: x)
    transforms[DateTime] = lambda x: x and dtparse(x) or None

    transform_lookup = {col.name: transforms[col.type.__class__] for col in model.__table__.c}

    def _transformer(content):
        try:
            return {k: transform_lookup[k](v) for k, v in content.items()}
        except ValueError as e:
            raise ParserError(f"Failed to parse row: {e}, at {content!r}") from e

    return _transformer


def create_meta_table_data(engine, dataset_schema: DatasetSchema):
    session = sessionmaker(bind=engine)()
    ds_content = {camel_case_to_snake(k): v for k, v in dataset_schema.items() if k != "tables"}
    ds_content["contact_point"] = str(ds_content.get("contact_point", ""))
    ds_transformer = transformer_factory(models.Dataset)
    dataset = models.Dataset(**ds_transformer(ds_content))
    session.add(dataset)

    for table_data in dataset_schema["tables"]:
        table_content = {camel_case_to_snake(k): v for k, v in table_data.items() if k != "schema"}

        table = models.Table(
            **{
                **table_content,
                **{f: table_data["schema"].get(f) for f in ("required", "display")},
            }
        )
        table.dataset_id = dataset.id
        session.add(table)

        for field_name, field_value in table_data["schema"]["properties"].items():
            field_content = {
                k.replace("$", ""): v for k, v in field_value.items() if k not in {"$comment"}
            }
            field_content["name"] = field_name
            try:
                field = models.Field(**field_content)
            except TypeError as e:
                raise NotImplementedError(
                    f'Import failed: at "{field_name}": {field_value!r}:\n{e}'
                ) from e

            field.table_id = table.id
            field.dataset_id = dataset.id
            session.add(field)

    session.commit()


def fetch_schema_from_relational_schema(engine, dataset_id) -> dict:
    """Restore the schema based on the stored metadata"""
    session = sessionmaker(bind=engine)()
    dataset = (
        session.query(models.Dataset)
        .join(models.Dataset.tables)
        .join(models.Table.fields)
        .filter(models.Dataset.id == dataset_id)
        .first()
    )
    if not dataset:
        raise ValueError(f"Dataset {dataset_id} not found.")

    aschema = _serialize(dataset)
    contact_point = aschema["contactPoint"]
    aschema["contactPoint"] = json.loads(contact_point)
    aschema["tables"] = [_serialize(t) for t in aschema["tables"]]
    for table_dict in aschema["tables"]:
        del table_dict["dataset"]
        del table_dict["datasetId"]
        table_dict["schema"] = {f: table_dict.get(f) for f in ("required", "display")}
        table_dict["schema"]["$schema"] = "http://json-schema.org/draft-07/schema#"
        table_dict["schema"]["type"] = "object"
        table_dict["schema"]["additionalProperties"] = False
        table_dict.pop("required", "")
        table_dict.pop("display", "")
        properties = [_serialize(f, camelize=False) for f in table_dict["fields"]]
        del table_dict["fields"]
        for prop in properties:
            del prop["table"]
            del prop["dataset_id"]
            del prop["table_id"]
            ref = prop.pop("ref", None)
            if ref is not None:
                prop["$ref"] = ref
            elif "geojson" in prop["type"]:
                prop["$ref"] = prop["type"]
                del prop["type"]
        schema_properties = _extract_names(properties)
        schema_version = table_dict.pop("schemaVersion")
        schema_properties["schema"] = {
            "$ref": (
                f"https://schemas.data.amsterdam.nl/schema@{schema_version}#/definitions/schema"
            )
        }
        table_dict["schema"]["properties"] = schema_properties
    return aschema


def _serialize(obj, camelize=True):
    results = {}
    for attr in inspect(obj).attrs:
        value = attr.value
        key = attr.key
        if camelize:
            key = snake_case_to_camel(key, upper_case_first=False)
        if value is None:
            continue
        if hasattr(value, "isoformat"):
            value = attr.value.isoformat()
        results[key] = value
    return results


def _extract_names(properties):
    return {toCamelCase(prop.pop("name")): prop for prop in properties}



import inspect
import textwrap
from argparse import ArgumentParser
from datetime import date, datetime
from typing import Any, Dict, List, Tuple, Type

from django.apps import AppConfig, apps
from django.core.management import BaseCommand
from django.db import models
from django.utils import timezone
from django.utils.functional import partition

from schematools.contrib.django.models import DynamicModel


class Command(BaseCommand):
    """Dump the dynamically generated models."""

    help = "Dump the (dynamic) model definitions that Django holds in-memory."  # noqa: A003

    model_meta_args: List[Tuple[str, Any]] = [
        # All possible options in Meta, with their defaults.
        # https://docs.djangoproject.com/en/3.2/ref/models/options/
        # The original model.Meta or Options.meta is not available after construction,
        # so will have to fill out the defaults to generate a reasonable representation.
        ("abstract", False),
        ("app_label", None),
        ("base_manager_name", None),
        ("constraints", []),
        ("db_table", None),
        ("default_manager_name", None),
        ("default_related_name", None),
        ("get_latest_by", None),
        ("index_together", ()),
        ("indexes", []),
        ("managed", True),
        ("order_with_respect_to", None),
        ("ordering", ["id"]),
        ("permissions", []),
        ("proxy", False),
        ("unique_together", ()),
        ("verbose_name", ""),  # is modelname
        ("verbose_name_plural", ""),  # is {modelname}s by default
    ]

    path_aliases: List[Tuple[str, str]] = [
        ("django.db.models.", "models."),
        ("django.contrib.gis.db.models.fields.", ""),
        ("django_postgres_unlimited_varchar.", ""),
    ]

    def add_arguments(self, parser: ArgumentParser) -> None:
        """Hook to add arguments."""
        parser.add_argument(
            "args", metavar="app_label", nargs="*", help="Names of Django apps to dump"
        )

    def handle(self, *args: str, **options: Any) -> None:
        """Main function of this command."""
        app_labels = sorted(args or apps.app_configs.keys())

        for app_label in app_labels:
            app: AppConfig = apps.get_app_config(app_label)
            self.write_header(app)
            models = app.get_models(include_auto_created=True)
            for model in sorted(models, key=lambda m: m._meta.model_name):
                self.write_model(model)

    def write_header(self, app: AppConfig) -> None:
        """Write app start header."""
        self.stdout.write(f"# ---- App: {app.verbose_name or app.label}\n\n\n")

    def write_model(self, model: Type[models.Model]) -> None:
        """Write the representation of a complete Django model to the output."""
        bases = ", ".join(base_class.__name__ for base_class in model.__bases__)
        self.stdout.write(f"class {model.__name__}({bases}):\n")

        if model.__doc__:
            self.write_docstring(model.__doc__)

        # Get model fields, split in auto created vs declared fields
        all_fields = model._meta.get_fields(include_parents=False, include_hidden=True)
        declared_fields, auto_created_fields = partition(lambda f: f.auto_created, all_fields)
        for field in declared_fields:
            self.write_field(field)

        if auto_created_fields:
            self.stdout.write("\n    # Auto created fields:\n")
            for field in sorted(auto_created_fields, key=lambda f: f.name):
                if isinstance(field, models.ForeignObjectRel):
                    self.write_reverse_field(field)
                else:
                    self.write_field(field)

        self.stdout.write("\n")

        if issubclass(model, DynamicModel):
            self.write_dynamic_model_attrs(model)

        self.write_model_meta(model)

        if issubclass(model, DynamicModel) and model._display_field:
            self.write_model_str(model)

        self.stdout.write("\n\n")

    def write_docstring(self, doc: str) -> None:
        """Write the docstring for a class."""
        if "\n" not in doc:  # if not formatted already
            # Wrap our description text
            doc = textwrap.fill(doc, 96, subsequent_indent="    ")
            if "\n" in doc:  # if wrapped
                doc += "\n"
        self.stdout.write(f'    """{doc}"""\n\n')

    def write_dynamic_model_attrs(self, model: Type[DynamicModel]) -> None:
        """Write the attributes defined by model_factory() for completeness."""
        self.stdout.write("    # Set by model_factory()\n")
        self.stdout.write(f"    # _dataset = {model._dataset!r}\n")
        self.stdout.write(f"    # _table_schema = {model._table_schema}\n")
        self.stdout.write(f"    _display_field = {model._display_field!r}\n")
        if model._is_temporal:
            self.stdout.write("    _is_temporal = True\n")
        self.stdout.write("\n")

    def write_model_meta(self, model: Type[models.Model]) -> None:
        """Write the 'class Meta' section."""
        self.stdout.write("    class Meta:\n")
        for meta_arg, default in self.model_meta_args:
            value = getattr(model._meta, meta_arg, ...)
            if value != default:
                self.stdout.write(f"        {meta_arg} = {value!r}\n")

    def write_model_str(self, model: Type[DynamicModel]) -> None:
        """For dynamic model, we know how __str__() looks like."""
        self.stdout.write("\n")
        self.stdout.write("    def __str__(self):\n")
        self.stdout.write(f"        return self.{model._display_field}\n")

    def write_field(self, field: models.Field) -> None:
        """Write how a field would have been written in a models file."""
        # Note that migration files use the 'name' from field.deconstruct()
        # but this is always identical to 'field.name' for standard Django fields.
        self.stdout.write(f"    {field.name} = {self._get_field_repr(field)}\n")

    def _get_field_repr(self, field: models.Field) -> str:
        """Generate the field representation, it's reused for arguments."""
        name, path, args, kwargs = field.deconstruct()
        return self._get_deconstructable_repr(path, args, kwargs)

    def _get_deconstructable_repr(self, path: str, args: List[Any], kwargs: Dict[str, Any]) -> str:
        for prefix, alias in self.path_aliases:
            if path.startswith(prefix):
                path = alias + path[len(prefix) :]

        str_args = ", ".join(self._format_value(arg) for arg in args)
        str_kwargs = ", ".join(f"{n}={self._format_value(v)}" for n, v in kwargs.items())
        return f"{path}({str_args}{str_kwargs})"

    def write_reverse_field(self, field: models.ForeignObjectRel) -> None:
        """Mention the reverse relations for clarity."""
        source_field = field.remote_field

        cls_name = field.__class__.__name__
        related_model = _format_model_name(field.related_model)
        source_name = f"<{source_field.__class__.__name__} at {related_model}.{source_field.name}>"

        comment = "hidden reverse relation" if field.is_hidden() else "reverse relation"
        self.stdout.write(
            f"    # {comment}: {field.name} = {cls_name}(field={source_name}, ...)\n"
        )

    def _format_value(self, value: Any) -> str:
        """Format the model kwarg, some callables should be mapped to their code name."""
        if isinstance(value, models.Field):
            # e.g. ArrayField(base_field=models.CharField(...))
            return self._get_field_repr(value)
        elif isinstance(value, type):
            return value.__name__
        elif isinstance(value, object) and hasattr(value, "deconstruct"):
            # A @deconstructible, e.g. URLPathValidator.
            return self._get_deconstructable_repr(*value.deconstruct())
        elif isinstance(value, list):
            # for validators=[...]
            return "[{0}]".format(",".join(self._format_value(v) for v in value))
        elif callable(value):
            if value is datetime.now:
                return "datetime.now"
            elif value is date.today:
                return "date.today"
            elif value is timezone.now:
                return "timezone.now"
            elif value is models.CASCADE:
                return "models.CASCADE"
            elif value is models.SET_NULL:
                return "models.SET_NULL"
            elif value is models.PROTECT:
                return "models.DO_NOTHING"
            elif value is models.SET_DEFAULT:
                return "models.SET_DEFAULT"
            elif value is models.PROTECT:
                return "models.PROTECT"
            elif inspect.isfunction(value):
                # e.g. validators=[some_function]
                return f"{value.__module__}.{value.__qualname__}"

        return repr(value)


def _format_model_name(model: Type[models.Model]) -> str:
    return f"{model._meta.app_label}.{model._meta.model_name}"



from django.apps import AppConfig


class SchematoolsAppConfig(AppConfig):
    name = "schematools.contrib.django"
    # Alias as `datasets` app to avoid unnecessary migrations on legacy systems.
    label = "datasets"



import copy

from sqlalchemy import inspect

from ..utils import toCamelCase
from .utils import DATASET_TMPL, TABLE_TMPL

# the Geometry field has a property geometry_type, could be mapped to more
# specific types in geojson.org

DB_TO_ASCHEMA_TYPE = {
    "DATE": {"type": "string", "format": "date"},
    "TIME": {"type": "string", "format": "time"},
    "TIMESTAMP": {"type": "string", "format": "date-time"},
    "VARCHAR": {"type": "string"},
    "INTEGER": {"type": "integer"},
    "BIGINT": {"type": "integer"},
    "SMALLINT": {"type": "integer"},
    "NUMERIC": {"type": "number"},
    "DOUBLE_PRECISION": {"type": "number"},
    "BOOLEAN": {"type": "boolean"},
    "TEXT": {"type": "string"},
    "ARRAY": {"type": "array"},
    "GEOMETRY": {"$ref": "https://geojson.org/schema/Geometry.json"},
    "POLYGON": {"$ref": "https://geojson.org/schema/Polygon.json"},
    "MULTIPOLYGON": {"$ref": "https://geojson.org/schema/MultiPolygon.json"},
    "POINT": {"$ref": "https://geojson.org/schema/Point.json"},
    "MULTIPOINT": {"$ref": "https://geojson.org/schema/MultiPoint.json"},
    "LINESTRING": {"$ref": "https://geojson.org/schema/LineString.json"},
    "MULTILINESTRING": {"$ref": "https://geojson.org/schema/MultiLineString.json"},
    "GEOMETRYCOLLECTION": {"$ref": "https://geojson.org/schema/GeometryCollection.json"},
}


def fix_name(field_name, field_value=None):
    ret = field_name
    if field_value is None or "relation" in field_value:
        ret = field_name.replace("_id", "")
    return toCamelCase(ret)


def introspect_db_schema(engine, dataset_id, tablenames, db_schema=None, prefix=None):
    """Generate an amsterdam schema file based on an existing database."""
    insp = inspect(engine)
    tables = []

    for full_table_name in tablenames:
        table_name = full_table_name
        if prefix is not None:
            table_name = full_table_name[len(prefix) :]
        columns = {}
        relations = {}
        required_field_names = ["schema"]
        fks = insp.get_foreign_keys(full_table_name, schema=db_schema)
        for fk in fks:
            if len(fk["constrained_columns"]) > 1:
                raise Exception("More than one fk col")
            constrained_column = fk["constrained_columns"][0]
            if not constrained_column.startswith("_"):
                relations[constrained_column] = fk["referred_table"]
        pk_info = insp.get_pk_constraint(full_table_name, schema=db_schema)
        if len(pk_info["constrained_columns"]) > 1:
            raise Exception("multicol pk")
        if len(pk_info["constrained_columns"]) == 0:
            raise Exception("no pk")
        # pk = pk_info["constrained_columns"][0]  # ASchema assumes 'id' for the pk
        for col in insp.get_columns(full_table_name, schema=db_schema):  # name, type, nullable
            col_name = col["name"]
            if col_name.startswith("_"):
                continue
            col_type = col["type"].__class__.__name__
            if col_type == "Geometry":
                col_type = col["type"].geometry_type
            if not col["nullable"]:
                required_field_names.append(col_name)
            aschema_type = DB_TO_ASCHEMA_TYPE[col_type].copy()
            columns[col_name] = aschema_type
            if col_type == "ARRAY":
                item_type = col["type"].item_type.__class__.__name__
                aschema_type["items"] = DB_TO_ASCHEMA_TYPE[item_type]
            # XXX Add 'title' and 'description' to the column

        for field_name, referred_table in relations.items():
            columns[field_name].update({"relation": referred_table.replace("_", ":", 1)})

        # Generate table section
        table = copy.deepcopy(TABLE_TMPL)
        table["id"] = table_name
        table["schema"]["required"] = [
            fix_name(fn, fv)
            for fn, fv in map(lambda n: (n, columns.get(n, None)), required_field_names)
        ]
        table["schema"]["properties"].update({fix_name(fn, fv): fv for fn, fv in columns.items()})
        tables.append(table)

    # Generate main section
    dataset = copy.deepcopy(DATASET_TMPL)
    dataset["id"] = dataset_id
    dataset["title"] = dataset_id
    dataset["tables"] = tables
    return dataset



"""Semantic JSON schema validation.

This package provides a simple :class:`Validator` class
that allows us to perform semantic validation
on a given JSON schema.

Semantic validation is different from structural validation,
as performed by the :mod:`jsonschema`,
in that it takes into account the implied logic behind a JSON schema.
For instance,
we might indicate that what identifies a table
is a set of properties (columns).
In Amsterdam Schema this is done using the ``identifier`` property
that can take an array of property names.
Structural validation cannot verify that the properties specified in that array
actually do exist in the schema.
Semantic validation can and should.

Specific semantic validations are created by subclassing :class:`Validator`
and overriding the :meth:`Validator.validate` method.
Simply by means of subclassing,
the new validator is automatically registered.
Registered validator classes will all be run
when :meth:`Validator.run_all()` is invoked.

.. note::

   For the registration to work
   all :class:`Validator` subclasses need to be parsed by the Python interpreter.
   This can be achieved by importing the module they reside in.

Example:
    The following will run all registered validators
    on a dataset ``dataset``.
    Any validation errors are printed to ``stdout``::

        dataset = _get_dataset_schema(meta_schema_url, schema_location)
        validator = Validator(dataset=dataset)
        for error in validator.run_all():
            print(error)

"""
from __future__ import annotations

import operator
import re
from dataclasses import dataclass
from functools import partial
from typing import Callable, ClassVar, Iterator, List, Set, Type, cast, final

from schematools import MAX_TABLE_NAME_LENGTH
from schematools.types import DatasetSchema
from schematools.utils import to_snake_case


@dataclass(frozen=True)
class ValidationError:
    """Capture validation errors."""

    validator_name: str
    message: str

    def __str__(self) -> str:
        return f"[{self.validator_name}] {self.message}"


class ValidationException(Exception):
    """Raised when validation fails to execute.

    .. note::

       This is not for validation errors. See :class:`ValidationError` instead.
    """

    message: str

    def __init__(self, message: str) -> None:
        super().__init__(message)
        self.message = message


class Validator:
    """Base class for validators.

    Not only is this a base class for validators,
    it is also used for running all registered validators.
    See Also: :meth:`run_all`

    Registration is a side-effect of overriding this base class.

    """

    _registry: ClassVar[List[Type[Validator]]] = []
    dataset: DatasetSchema

    @classmethod
    def __init_subclass__(cls: Type[Validator]) -> None:
        """Register sub classes."""
        super().__init_subclass__()
        cls._registry.append(cls)

    def __init__(self, dataset: DatasetSchema) -> None:
        """Initialize the validator with a dataset.

        Args:
            dataset: The dataset to run the validations on.
        """
        self.dataset = dataset

    def validate(self) -> Iterator[ValidationError]:
        """Run validation."""
        raise NotImplementedError(
            f"{self.__class__.__name__}.{self.validate.__name__} should be overridden in "
            f"subclasses and called from there."
        )

    @final
    def run_all(self) -> Iterator[ValidationError]:
        """Run all registered validators.

        Yields:
            :class:`ValidationError`\s if any.

        """  # noqa: W605
        for validator_cls in self._registry:
            validator_inst = validator_cls(dataset=self.dataset)
            yield from validator_inst.validate()


class PsqlIdentifierLengthValidator(Validator):
    """Validate inferred PostgreSQL table names for not exceeding max length.

    PostgreSQL has a maximum length for identifiers such as table names.
    We infer table names from dataset and table ids in the schemas.
    Those inferred table names should not exceed the max identifier length
    supported by PostgreSQL.
    """

    def validate(self) -> Iterator[ValidationError]:
        for table in self.dataset.get_tables(include_nested=True, include_through=True):
            # `table_name` should probably be a property on `DatasetTableSchema`.
            # There already is a `db_name` property however it truncates the inferred
            # table name to prevent exceeding the PostgreSQL limit. Hence it cannot
            # be used for validation purposes.
            table_name = to_snake_case("_".join((self.dataset.id, table.name)))
            if (length := len(table_name)) > MAX_TABLE_NAME_LENGTH:
                excess = length - MAX_TABLE_NAME_LENGTH
                yield ValidationError(
                    self.__class__.__name__,
                    f"Inferred PostgreSQL table name '{table_name}' is '{excess}' characters too "
                    f"long. Maximum table name length is '{MAX_TABLE_NAME_LENGTH}'.",
                )


class IdentPropRefsValidator(Validator):
    """Validate that the identifier property refers to actual fields on the table definitions."""

    def validate(self) -> Iterator[ValidationError]:
        @dataclass
        class DerivedField:
            original: str
            derived: str

        for table in self.dataset.get_tables(include_nested=True):
            identifiers = set(table.identifier)
            table_fields = cast(Set[str], set(map(operator.attrgetter("id"), table.fields)))
            if not identifiers.issubset(table_fields):
                missing_fields = identifiers - table_fields
                # The 'identifier' property is weird in that it is not exclusively defined in
                # terms of literally defined fields on the table. For instance, given a relation:
                #
                #     "indicatorDefinitie": {
                #       "type": "string",
                #       "relation": "bbga:indicatorenDefinities",
                #        "description": "De variabele in kwestie."
                #     }
                #
                # 'identifier' can refer to this field as 'identifierDefinitionId' (mind the
                # 'Id' postfix). Simply referring to this field (from 'identifier') as
                # 'indicatorDefinitie', eg as:
                #
                #     "identifier": ["indicatorDefinitie", "jaar", "gebiedcode15"],
                #
                #  will NOT work. It has to be postfixed with 'Id', eg:
                #
                #     "identifier": ["indicatorDefinitieId", "jaar", "gebiedcode15"],
                #
                # I think this is a bug is schema-tools, but for now I'll cover this case
                # explicitly.
                remove_id_suffix = cast(Callable[[str], str], partial(re.sub, r"(.+)Id", r"\1"))
                derived_fields = tuple(
                    map(
                        lambda f: DerivedField(original=remove_id_suffix(f), derived=f),
                        missing_fields,
                    )
                )
                for df in derived_fields:
                    if df.original in table_fields:
                        missing_fields.discard(df.derived)
                if missing_fields:
                    fields, have = (
                        ("fields", "have") if len(missing_fields) > 1 else ("field", "has")
                    )
                    yield ValidationError(
                        self.__class__.__name__,
                        f"Property 'identifier' on table '{table.id}' refers to {fields} "
                        f"'{', '.join(missing_fields)}' that {have} not been defined on the "
                        "table.",
                    )



import hashlib
import operator
from collections import Counter, UserDict
from functools import reduce
from itertools import islice
from logging import Logger
from pathlib import PosixPath
from typing import Any, Dict, Iterator, List, Optional, Set, Union, cast

from jsonpath_rw import parse
from jsonpath_rw.jsonpath import Child
from sqlalchemy import exc, inspect
from sqlalchemy.dialects.postgresql.base import PGInspector
from sqlalchemy.engine.base import Engine
from sqlalchemy.sql.schema import Index, MetaData, Table

from schematools import MAX_TABLE_NAME_LENGTH, TABLE_INDEX_POSTFIX
from schematools.factories import tables_factory
from schematools.types import DatasetSchema, DatasetTableSchema
from schematools.utils import to_snake_case, toCamelCase

metadata = MetaData()


def chunked(generator: Iterator[Any], size: int) -> Iterator[Any]:
    """Read parts of the generator, pause each time after a chunk."""
    # Based on more-itertools. islice returns results until 'size',
    # iter() repeatedly calls make_chunk until the '[]' sentinel is returned.
    gen = iter(generator)
    make_chunk = lambda: list(islice(gen, size))
    return iter(make_chunk, [])


class JsonPathException(Exception):
    """Exception used to signal a jsonpath provenance."""

    pass


class Row(UserDict):
    """Dict-based class that used provenance to find values."""

    # class-level cache for jsonpath expressions
    _expr_cache: Dict[str, Child] = {}

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        """Initializer that sets the provenance information."""
        fields_provenances = kwargs.pop("fields_provenances", {})
        self.rev_provenances: Dict[str, str] = {
            name: prov_name for prov_name, name in fields_provenances.items()
        }
        self.field_provenanced_by_id: Optional[str] = fields_provenances.get("id")
        # Provenanced keys are stored in a cache. This is not only for efficiency.
        # Sometimes, the key that is 'provenanced' is the same key that is in the ndjson
        # import data. When this key gets replaced, the original object structure
        # is not available anymore, so subsequent jsonpath lookups will fail.
        self.provenances_cache: Dict[str, str] = {}

        super().__init__(*args, **kwargs)

        if self.field_provenanced_by_id is not None:
            self.data[self.field_provenanced_by_id] = self.data["id"]

    def __getitem__(self, key: str) -> Any:
        """Gets a value taking provenance into account."""
        try:
            value = super().__getitem__(self._transform_key(key))
        except JsonPathException:
            value = self.provenances_cache.get(key) or self._fetch_value_for_jsonpath(key)
            self.provenances_cache[key] = value
        return value

    def __delitem__(self, key: str) -> None:
        """Deletes a value taking provenance into account."""
        try:
            return super().__delitem__(self._transform_key(key))
        except JsonPathException:
            # For JsonPath keys, a __del__ should use the
            # key that is originally provided
            return super().__delitem__(key)

    def _transform_key(self, key: str) -> str:
        if key == self.field_provenanced_by_id:
            return self.field_provenanced_by_id
        prov_key = self.rev_provenances.get(key)
        if prov_key is not None:
            if prov_key.startswith("$"):
                raise JsonPathException()
            return prov_key
        if key in self.data:
            return key
        raise KeyError

    def _fetch_expr(self, prov_key: str) -> Child:
        if prov_key in self._expr_cache:
            return self._expr_cache[prov_key]
        expr = parse(prov_key)
        self._expr_cache[prov_key] = expr
        return expr

    def _fetch_value_for_jsonpath(self, key: str) -> Optional[Union[int, str]]:
        prov_key = self.rev_provenances.get(key)
        if prov_key is None:
            return None
        top_element_name = prov_key.split(".")[1]
        expr = self._fetch_expr(prov_key)
        # Sometimes the data is not an object, but simply None
        top_level_data = self.data[top_element_name]
        if top_level_data is None:
            self.data[key] = None
            return None
        matches = expr.find({top_element_name: top_level_data})
        value = None if not matches else matches[0].value
        self.data[key] = value
        return value


class BaseImporter:
    """Base importer that holds common data."""

    def __init__(
        self, dataset_schema: DatasetSchema, engine: Engine, logger: Optional[Logger] = None
    ) -> None:
        """Initializes the BaseImporter.

        dataset_schema: The dataset to work with.
        engine: SQLAlchemy database engine.
        """
        self.engine = engine
        self.dataset_schema = dataset_schema
        self.srid = dataset_schema["crs"].split(":")[-1]
        self.dataset_table: Optional[DatasetTableSchema] = None
        self.fields_provenances: Dict[str, str] = {}
        self.db_table_name: Optional[str] = None
        self.tables: Dict[str, Table] = {}
        self.indexes: Dict[str, List[Index]] = {}
        self.pk_values_lookup: Dict[str, Set[Any]] = {}
        self.pk_colname_lookup: Dict[str, str] = {}
        self.logger = LogfileLogger(logger) if logger else CliLogger()

    def fetch_fields_provenances(self, dataset_table: DatasetTableSchema) -> Dict[str, str]:
        """Create mapping from provenance to camelcased fieldname."""
        fields_provenances = {}
        for field in dataset_table.fields:
            if (provenance := field.get("provenance")) is not None:
                fields_provenances[provenance] = field.name
        return fields_provenances

    def fix_fieldnames(
        self, fields_provenances: Dict[str, str], table_records: Iterator[Any]
    ) -> Any:
        """Fixes the fieldname.

        We need relational snakecased fieldnames in the records and,
        we need to take provenance in the input records into account.
        """
        fixed_records = []
        for record in table_records:
            fixed_record = {}
            for field_name, field_value in record.items():
                if field_name == "id":
                    fixed_record["id"] = field_value
                    continue
                fixed_field_name = fields_provenances.get(field_name, field_name)
                fixed_record[to_snake_case(fixed_field_name)] = field_value
            fixed_records.append(fixed_record)
        return fixed_records

    def deduplicate(
        self,
        table_name: str,
        table_records: List[Row],
    ) -> Iterator[Row]:
        """Removes duplicates from a set of records."""
        this_batch_pk_values = set()
        pk_name = self.pk_colname_lookup.get(table_name)

        values_lookup: Set[Any] = self.pk_values_lookup.get(table_name, set())

        for record in table_records:
            if pk_name is None:
                yield record
                continue
            value = record[toCamelCase(pk_name)]
            if value not in values_lookup and value not in this_batch_pk_values:
                yield record
            else:
                self.logger.log_warning(
                    "Duplicate record for %s, with %s = %s", table_name, pk_name, value
                )
            this_batch_pk_values.add(value)
            values_lookup.add(value)

    def create_pk_lookup(self, tables: Dict[str, Table]) -> None:
        """Generate a lookup to avoid primary_key clashes."""
        for table_name, table in tables.items():
            if isinstance(table, Table):
                pk_columns = inspect(table).primary_key.columns
                # nm tables do not have a PK
                if not pk_columns:
                    continue
                # We assume a single PK (because of Django)
                pk_col = pk_columns.values()[0]
                pk_name = pk_col.name
                if pk_col.autoincrement:
                    continue
                self.pk_colname_lookup[table_name] = pk_name
                pks = {getattr(r, pk_name) for r in self.engine.execute(table.select())}

                self.pk_values_lookup[table_name] = pks

    def generate_db_objects(
        self,
        table_id: str,
        db_schema_name: Optional[str] = None,
        db_table_name: Optional[str] = None,
        truncate: bool = False,
        ind_tables: bool = True,
        ind_extra_index: bool = True,
        limit_tables_to: Optional[Set] = None,
    ) -> None:
        """Generate the tablemodels, tables and indexes.

        As default both table and index creation are set to True.

        Args:
            table_id: Name of the table as defined in the JSON schema defintion.
            db_schema_name: Name of the database schema where table should be
                created/present. Defaults to None (== public).
            db_table_name: Name of the table as defined in the database.
                Defaults to None.
            truncate: Indication to truncate table. Defaults to False.
            ind_tables: Indication to create table. Defaults to True.
            ind_extra_index: Indication to create indexes. Defaults to True.
            limit_tables_to: Only process the indicated tables. Normally, SA tables
                are generated for the whole dataset where `table_id` belongs to.
                Sometimes, this is not needed/wanted.
        """
        self.dataset_table = cast(
            DatasetTableSchema, self.dataset_schema.get_table_by_id(table_id)
        )
        if ind_tables or ind_extra_index:

            # Collect provenance info for easy re-use
            self.fields_provenances = self.fetch_fields_provenances(self.dataset_table)
            self.db_table_name = db_table_name
            # Bind the metadata
            metadata.bind = self.engine
            # Get a table to import into

            if (dataset := self.dataset_table.dataset) is None:
                raise ValueError("Table {table_id} does not belong to a dataset")

            self.tables = tables_factory(
                dataset,
                metadata=metadata,
                db_schema_names={table_id: db_schema_name},
                db_table_names={table_id: db_table_name},
                limit_tables_to=limit_tables_to,
            )

        if db_table_name is None:
            self.db_table_name = self.dataset_table.db_name()

        if ind_tables:
            self.prepare_tables(self.tables, truncate=truncate)
            self.create_pk_lookup(self.tables)

        if ind_extra_index:
            try:
                # Get indexes to create
                self.indexes = index_factory(
                    self.dataset_table,
                    ind_extra_index,
                    db_schema_name=db_schema_name,
                    metadata=metadata,
                    db_table_name=self.db_table_name,
                )
                metadata_inspector = inspect(metadata.bind)
                self.prepare_extra_index(
                    self.indexes,
                    metadata_inspector,
                    metadata.bind,
                    db_schema_name,
                )
            except exc.NoInspectionAvailable:
                pass

    def load_file(
        self,
        file_name: PosixPath,
        batch_size: int = 100,
        is_through_table: bool = False,
        **kwargs: Any,
    ) -> Optional[Row]:
        """Import a file into the database table, returns the last record, if available."""
        if self.dataset_table is None:
            raise ValueError("Import needs to be initialized with table info")
        data_generator = self.parse_records(
            file_name,
            self.dataset_table,
            self.db_table_name,
            is_through_table=is_through_table,
            **{"fields_provenances": self.fields_provenances, **kwargs},
        )
        self.logger.log_start(file_name, size=batch_size)

        num_imported = 0
        insert_statements = {
            table_name: table.insert() for table_name, table in self.tables.items()
        }

        last_record: Optional[Row] = None
        for records in chunked(data_generator, size=batch_size):
            # every record is keyed on tablename + inside there is a list
            for table_name, insert_statement in insert_statements.items():
                table_records: List[Row] = reduce(
                    operator.add, [record.get(table_name, []) for record in records], []
                )
                table_records = self.fix_fieldnames(
                    self.fields_provenances,
                    self.deduplicate(table_name, table_records),
                )
                if table_records:
                    self.engine.execute(
                        insert_statement,
                        table_records,
                    )
            num_imported += len(records)
            self.logger.log_progress(num_imported)
            # main table is keyed on tablename and only has one row
            last_record = records[-1][self.dataset_table.name][0]

        self.logger.log_done(num_imported)
        return last_record

    def parse_records(
        self,
        filename: PosixPath,
        dataset_table: DatasetTableSchema,
        db_table_name: Optional[str] = None,
        is_through_table: bool = False,
        **kwargs: Any,
    ) -> Iterator[Dict[str, List[Row]]]:
        """Yield all records from the filename."""
        raise NotImplementedError()

    def prepare_tables(self, tables: Dict[str, Table], truncate: bool = False) -> None:
        """Create the tables if needed."""
        for table in tables.values():
            if isinstance(table, Table):
                if not table.exists():
                    table.create()
                elif truncate:
                    self.engine.execute(table.delete())

    def prepare_extra_index(
        self,
        indexes: Dict[str, List[Index]],
        inspector: PGInspector,
        engine: Engine,
        db_schema_name: Optional[str] = None,
        logger: Optional[Logger] = None,
    ) -> None:
        """Create extra indexes.

        Indexs are placed on identifiers columns in base tables
        and identifier columns in n:m tables, if not exists.
        """
        # setup logger
        _logger: CliLogger = LogfileLogger(logger) if logger else CliLogger()

        # In the indexes dict, the table name of each index, is stored in the key
        target_table_names = list(indexes.keys())
        missing_table_names = set()

        # get current DB indexes on table
        current_db_indexes = set()
        for table in target_table_names:
            try:
                db_indexes = inspector.get_indexes(table, schema=db_schema_name)
            except exc.NoSuchTableError:
                missing_table_names.add(table)
                continue
            for current_db_index in db_indexes:
                current_db_indexes.add(current_db_index["name"])

        # get all found indexes generated out of Amsterdam schema
        schema_indexes = set()
        schema_indexes_objects = {}
        for index_object_list in indexes.values():
            for index_object in index_object_list:
                schema_indexes.add(index_object.name)
                schema_indexes_objects[index_object.name] = index_object

        # get difference between DB indexes en indexes found Amsterdam schema
        indexes_to_create = list(
            (Counter(schema_indexes) - Counter(current_db_indexes)).elements()
        )

        # create indexes - that do not exists yet- in DB
        for index_name in indexes_to_create:
            try:
                index = schema_indexes_objects[index_name]
                if index.table.name not in missing_table_names:
                    _logger.log_warning(f"Index '{index_name}' not found...creating")
                    index.create(bind=engine)

            except AttributeError as e:
                _logger.log_error(
                    f"Error creating index '{index_name}' for '{target_table_names}', error: {e}"
                )
                continue


class CliLogger:
    """Logger to be used when importer is called from the cli."""

    def log_start(self, file_name: PosixPath, size: int) -> None:
        """Start the logging."""
        print(f"Importing data [each dot is {size} records]: ", end="", flush=True)  # noqa: T001

    def log_progress(self, num_imported: int) -> None:
        """Output progress."""
        print(".", end="", flush=True)  # noqa: T001

    def log_error(self, msg: str, *args: Any) -> None:
        """Output error."""
        print(msg % args)  # noqa: T001

    def log_warning(self, msg: str, *args: Any) -> None:
        """Output warning."""
        print(msg % args)  # noqa: T001

    def log_done(self, num_imported: int) -> None:
        """Indicate logging is finished."""
        print(f" Done importing {num_imported} records", flush=True)  # noqa: T001


class LogfileLogger(CliLogger):
    """Logger to be used when importer is called from python code."""

    def __init__(self, logger: Logger):
        """Initialize logger."""
        self.logger = logger

    def log_start(self, file_name: PosixPath, size: int) -> None:
        """Start the logging."""
        self.logger.info("Importing %s with %d records each:", file_name, size)

    def log_progress(self, num_imported: int) -> None:
        """Output progress."""
        self.logger.info("- imported %d records", num_imported)

    def log_error(self, msg: str, *args: Any) -> None:
        """Output error."""
        self.logger.error(msg, *args)

    def log_warning(self, msg: str, *args: Any) -> None:
        """Output warning."""
        self.logger.warning(msg, *args)

    def log_done(self, num_imported: int) -> None:
        """Indicate logging is finished."""
        self.logger.info("Done")


def index_factory(
    dataset_table: DatasetTableSchema,
    ind_extra_index: bool,
    metadata: Optional[MetaData] = None,
    db_table_name: Optional[str] = None,
    db_schema_name: Optional[str] = None,
    logger: Optional[Logger] = None,
) -> Dict[str, List[Index]]:
    """Generates one or more SQLAlchemy Index objects to work with the JSON Schema.

    :param dataset_table: The Amsterdam Schema definition of the table
    :param metadata: SQLAlchemy schema metadata that groups all tables to a single connection.
    :param db_table_name: Optional table name, which is otherwise inferred from the schema name.
    :param db_schema_name: Optional database schema name, which is otherwise None
        and defaults to public.

    Identifier index:
    In the JSON Schema definition of the table, an identifier arry may be definied.
    I.e. "identifier": ["identificatie", "volgnummer"]
    This is frequently used in temporal tables where versions of data are present (history).

    Through table index:
    In the JSON schema definition of the table, relations may be definied.
    In case of temporal data, this will lead to intersection tables a.k.a. through tables to
    accomodate n:m relations.

    FK index:
    In the JSON schema definition of the table, relations may be 1:N relations definied.
    Where the child table is referencing a parent table.
    These reference are not enforced by a database foreign key constraint. But will be
    used in joins to collect data when calling a API endpoint. Therefore must be indexed
    for optimal performance.

    The returned Index objects are keyed on the name of table
    """
    indexes = {}
    _metadata = cast(MetaData, metadata or MetaData())
    _logger = LogfileLogger(logger) if logger else CliLogger()

    # This is really ugly, mypy forces a new local variable here,
    # otherwise it keeps complaining about the type of db_table_name being Optional.
    if db_table_name is None:
        _db_table_name = dataset_table.db_name()
    else:
        _db_table_name = db_table_name

    table_name = f"{db_schema_name}.{_db_table_name}" if db_schema_name else _db_table_name

    try:
        table_object = _metadata.tables[table_name]
    except KeyError as ex:
        _logger.log_error(f"{table_name} cannot be found.", repr(ex))

    indexes_to_create = []

    def make_hash_value(index_name: str) -> str:
        """Create a hash value for index_name.

        Postgres DB holds currently 63 max characters for objectnames.
        To prevent exceeds and collisions, the index names are shortend
        based upon a hash.
        With the blake2s algorithm a digest size is set to 20 bytes,
        which proceduces a 40 character long heximal string plus
        the additional 4 character postfix of '_idx' (TABLE_INDEX_POSTFIX).
        """
        return (
            hashlib.blake2s(index_name.encode(), digest_size=20).hexdigest() + TABLE_INDEX_POSTFIX
        )

    def define_fk_index() -> None:
        """Creates an index on Foreign Keys."""
        if dataset_table.get_fk_fields():

            for field in dataset_table.get_fk_fields():
                field_name = f"{to_snake_case(field)}_id"
                index_name = f"{_db_table_name}_{field_name}_idx"
                if len(index_name) > MAX_TABLE_NAME_LENGTH:
                    index_name = make_hash_value(index_name)
                indexes_to_create.append(Index(index_name, table_object.c[field_name]))

        # add Index objects to create
        indexes[_db_table_name] = indexes_to_create

    def define_identifier_index() -> None:
        """Creates index based on the 'identifier' specification in the Amsterdam schema."""
        identifier_column_snaked = []

        if not indexes_to_create and dataset_table.identifier:

            for identifier_column in dataset_table.identifier:
                try:
                    identifier_column_snaked.append(
                        # precautionary measure: If camelCase, translate to
                        # snake_case, so column(s) can be found in table.
                        table_object.c[to_snake_case(identifier_column)]
                    )
                except KeyError as e:
                    _logger.log_error(f"{e.__str__} on {dataset_table.id}.{identifier_column}")
                    continue

        index_name = f"{_db_table_name}_identifier_idx"
        if len(index_name) > MAX_TABLE_NAME_LENGTH:
            index_name = make_hash_value(index_name)
        indexes_to_create.append(Index(index_name, *identifier_column_snaked))

        # add Index objects to create
        indexes[_db_table_name] = indexes_to_create

    def define_throughtable_index() -> None:
        """Creates index(es) on the many-to-many tables.

        Those are based on 'relation' specification in the Amsterdam schema.
        """
        for table in dataset_table.get_through_tables_by_id():

            through_columns: List[str] = []
            indexes_to_create: List[Index] = []

            # make a dictionary of the indexes to create
            if table.is_through_table:
                through_columns = []

                # First collect the fields that are relations
                relation_field_names = []
                for field in table.fields:
                    if field.relation:
                        snakecased_fieldname = to_snake_case(field.name)
                        relation_field_names.append(snakecased_fieldname)
                        through_columns.append(f"{snakecased_fieldname}_id")

                # Now check complementary field if they start with
                # the name of the relation field
                for field in table.fields:
                    if not field.relation:
                        snakecased_fieldname = to_snake_case(field.name)
                        if snakecased_fieldname.startswith(tuple(relation_field_names)):
                            through_columns.append(snakecased_fieldname)

            # create the Index objects
            if through_columns:
                table_id = table.db_name()

                try:
                    table_object = _metadata.tables[table_id]

                except KeyError:
                    _logger.log_error(
                        f"Unable to create Indexes {table_id}. Table not found in DB."
                    )
                    continue

                for column in through_columns:
                    index_name = table_id + "_" + column + TABLE_INDEX_POSTFIX
                    if len(index_name) > MAX_TABLE_NAME_LENGTH:
                        index_name = make_hash_value(index_name)
                    try:
                        indexes_to_create.append(Index(index_name, table_object.c[column]))
                    except KeyError as e:
                        _logger.log_error(
                            f"{e.__str__}:{table_id}.{column} not found in {table_object.c}"
                        )
                        continue

                # add Index objects to create
                indexes[table_id] = indexes_to_create

    if ind_extra_index:
        define_identifier_index()
        define_throughtable_index()
        define_fk_index()

    return indexes



"""Create GRANT statements to give roles very specific access to the database."""
from pg_grant import PgObjectType, parse_acl_item, query
from pg_grant.sql import grant, revoke
from sqlalchemy import text
from sqlalchemy.orm import sessionmaker

from schematools.types import DatasetSchema
from schematools.utils import to_snake_case

PUBLIC_SCOPE = "OPENBAAR"

existing_roles = set()


def introspect_permissions(engine, role):
    schema_relation_infolist = query.get_all_table_acls(engine, schema="public")
    for schema_relation_info in schema_relation_infolist:
        if schema_relation_info.acl:
            acl_list = [parse_acl_item(item) for item in schema_relation_info.acl]
            for acl in acl_list:
                if acl.grantee == role:
                    print(
                        'role "{}" has priviliges {} on table "{}"'.format(
                            role, ",".join(acl.privs), schema_relation_info.name
                        )
                    )


def revoke_permissions(engine, role):
    grantee = role
    schema_relation_infolist = query.get_all_table_acls(engine, schema="public")
    for schema_relation_info in schema_relation_infolist:
        if schema_relation_info.acl:
            acl_list = [parse_acl_item(item) for item in schema_relation_info.acl]
            for acl in acl_list:
                if acl.grantee == role:
                    print(
                        'revoking ALL priviliges of role "{}" on table "{}"'.format(
                            role, schema_relation_info.name
                        )
                    )
                    revoke_statement = revoke(
                        "ALL", PgObjectType.TABLE, schema_relation_info.name, grantee
                    )
                    engine.execute(revoke_statement)


def apply_schema_and_profile_permissions(
    engine,
    pg_schema,
    ams_schema,
    profiles,
    role,
    scope,
    set_read_permissions=True,
    set_write_permissions=True,
    dry_run=False,
    create_roles=False,
    revoke=False,
):
    Session = sessionmaker(bind=engine)
    session = Session()
    try:
        if ams_schema:
            create_acl_from_schemas(
                session,
                pg_schema,
                ams_schema,
                role,
                scope,
                set_read_permissions,
                set_write_permissions,
                dry_run,
                create_roles,
                revoke,
            )
        if profiles:
            profile_list = profiles.values()
            create_acl_from_profiles(engine, pg_schema, profile_list, role, scope)
        session.commit()
    except Exception:
        session.rollback()
        print("warning: session rolled back")
        raise
    finally:
        session.close()


def create_acl_from_profiles(engine, pg_schema, profile_list, role, scope):
    # NOTE: Rudimentary, not ready for production.
    acl_list = query.get_all_table_acls(engine, schema=pg_schema)
    priviliges = [
        "SELECT",
    ]
    grantee = role
    for profile in profile_list:
        if scope in profile["scopes"]:
            for dataset, _details in profile["schema_data"]["datasets"].items():
                for item in acl_list:
                    if item.name.startswith(dataset + "_"):
                        grant_statement = grant(
                            priviliges,
                            PgObjectType.TABLE,
                            item.name,
                            grantee,
                            grant_option=False,
                            schema=pg_schema,
                        )
                        print(grant_statement)
                        engine.execute(grant_statement)


def set_dataset_write_permissions(session, pg_schema, ams_schema, dry_run, create_roles):
    grantee = f"write_{ams_schema.id}"
    if create_roles:
        _create_role_if_not_exists(session, grantee, dry_run=dry_run)
    for table in ams_schema.get_tables(include_nested=True, include_through=True):
        table_name = "{}_{}".format(
            table.dataset.id, to_snake_case(table.id)
        )  # een aantal table.id's zijn camelcase
        table_privileges = ["INSERT", "UPDATE", "DELETE", "TRUNCATE", "REFERENCES"]
        _execute_grant(
            session,
            grant(
                table_privileges,
                PgObjectType.TABLE,
                table_name,
                grantee,
                grant_option=False,
                schema=pg_schema,
            ),
            dry_run=dry_run,
        )


def set_dataset_read_permissions(
    session, pg_schema, ams_schema, role, scope, dry_run, create_roles
):
    grantee = None if role == "AUTO" else role
    if create_roles and grantee:
        _create_role_if_not_exists(session, grantee)
    dataset_scope = (
        ams_schema.auth
        if ams_schema.auth
        else {
            PUBLIC_SCOPE,
        }
    )
    dataset_scope_set = {dataset_scope} if isinstance(dataset_scope, str) else set(dataset_scope)
    if dataset_scope_set - {PUBLIC_SCOPE}:
        print(
            'Found dataset read permission for "{}" to scopes "{}"'.format(
                ams_schema.id, dataset_scope_set
            )
        )

    for table in ams_schema.get_tables(include_nested=True, include_through=True):
        table_name = "{}_{}".format(
            table.dataset.id, to_snake_case(table.id)
        )  # een aantal table.id's zijn camelcase
        table_scope = table.auth if table.auth else dataset_scope
        table_scope_set = {table_scope} if isinstance(table_scope, str) else set(table_scope)
        if table.auth:
            print(f'Found table read permission for "{table_name}" to scopes "{table_scope_set}"')
            print(
                f'"{table_scope_set}" overrules "{dataset_scope_set}"'
                f' for read permission of "{table_name}"'
            )
        contains_field_grants = False
        fields = [field for field in table.fields if field.name != "schema"]
        for field in fields:
            if field.auth:
                field_scope = field.auth
                field_scope_set = (
                    {field_scope} if isinstance(field_scope, str) else set(field_scope)
                )
                print(
                    f'Found field read permission for "{field.name}" in'
                    f' table "{table_name}" for scopes {field_scope_set}'
                )
                contains_field_grants = True
                print(
                    f'"{field_scope_set}" overrules "{table_scope_set}" for read'
                    f' permission of field {field.name} in table {table_name}"'
                )
                if role == "AUTO":
                    grantees = [scope_to_role(scope) for scope in field_scope_set]
                elif scope in field_scope_set:
                    grantees = [role]
                else:
                    grantees = []
                for grantee in grantees:
                    if create_roles:
                        _create_role_if_not_exists(session, grantee, dry_run=dry_run)
                    column_name = to_snake_case(field.name)
                    # the space after SELECT is very important
                    column_priviliges = [f"SELECT ({column_name})"]
                    _execute_grant(
                        session,
                        grant(
                            column_priviliges,
                            PgObjectType.TABLE,
                            table_name,
                            grantee,
                            grant_option=False,
                            schema=pg_schema,
                        ),
                        dry_run=dry_run,
                    )
        if role == "AUTO":
            grantees = [scope_to_role(scope) for scope in table_scope_set]
        elif scope in table_scope_set:
            grantees = [role]
        else:
            grantees = []

        if contains_field_grants:
            # Only grant those fields without their own scope.
            # The other field have already been granted above
            for grantee in grantees:
                if create_roles:
                    _create_role_if_not_exists(session, grantee, dry_run=dry_run)
                for field in fields:
                    if not field.auth:
                        column_name = to_snake_case(field.name)
                        # the space after SELECT is very important:
                        column_priviliges = ["SELECT ({})".format(column_name)]
                        _execute_grant(
                            session,
                            grant(
                                column_priviliges,
                                PgObjectType.TABLE,
                                table_name,
                                grantee,
                                grant_option=False,
                                schema=pg_schema,
                            ),
                            dry_run=dry_run,
                        )
        else:
            # we can grant the whole table instead of field by field
            for grantee in grantees:
                if create_roles:
                    _create_role_if_not_exists(session, grantee, dry_run=dry_run)
                table_privileges = ["SELECT"]
                _execute_grant(
                    session,
                    grant(
                        table_privileges,
                        PgObjectType.TABLE,
                        table_name,
                        grantee,
                        grant_option=False,
                        schema=pg_schema,
                    ),
                    dry_run=dry_run,
                )


def create_acl_from_schemas(
    session,
    pg_schema,
    schemas,
    role,
    scopes,
    set_read_permissions,
    set_write_permissions,
    dry_run,
    create_roles,
    revoke,
):
    """Create and set the ACL for automatically generated roles based on Amsterdam Schema.
    Read permissions are granted to roles 'scope_X', where X are scopes found in Amsterdam Schema
    Write permissions are granted to roles 'write_Y', where Y are dataset ids,
    for all tables belonging to the dataset.
    Revoke old privileges before assigning new in case new privileges are more restrictive.
    """

    if revoke:
        if role == "AUTO":
            if isinstance(schemas, DatasetSchema):
                # for a single dataset
                _revoke_all_privileges_from_read_and_write_roles(
                    session, pg_schema, schemas, dry_run=dry_run
                )
            else:
                _revoke_all_privileges_from_read_and_write_roles(
                    session, pg_schema, dry_run=dry_run
                )
        else:
            if isinstance(schemas, DatasetSchema):
                # for a single dataset
                _revoke_all_privileges_from_role(
                    session, pg_schema, role, schemas, dry_run=dry_run
                )
            else:
                _revoke_all_privileges_from_role(session, pg_schema, role, dry_run=dry_run)

    if set_read_permissions:
        if isinstance(schemas, DatasetSchema):
            # for a single dataset
            set_dataset_read_permissions(
                session, pg_schema, schemas, role, scopes, dry_run, create_roles
            )
        else:
            for _dataset_name, dataset_schema in schemas.items():
                set_dataset_read_permissions(
                    session, pg_schema, dataset_schema, role, scopes, dry_run, create_roles
                )

    if set_write_permissions:
        if isinstance(schemas, DatasetSchema):
            # for a single dataset
            set_dataset_write_permissions(session, pg_schema, schemas, dry_run, create_roles)
        else:
            for _dataset_name, dataset_schema in schemas.items():
                set_dataset_write_permissions(
                    session, pg_schema, dataset_schema, dry_run, create_roles
                )


def _revoke_all_privileges_from_role(
    session, pg_schema, role, dataset_name=None, echo=True, dry_run=False
):
    status_msg = "Skipped" if dry_run else "Executed"
    if dataset_name:
        # for a single dataset
        revoke_statements = []
        for table in dataset_name.tables:
            revoke_statements.append(
                f"REVOKE ALL PRIVILEGES ON {pg_schema}.{table.db_name()} FROM {role}"
            )
        revoke_statement = ";".join(revoke_statements)
    else:
        revoke_statement = f"REVOKE ALL PRIVILEGES ON ALL TABLES IN {pg_schema} FROM {role}"
    sql_statement = (
        "DO $$ "
        "BEGIN "
        f"{revoke_statement}; "
        "EXCEPTION"
        " WHEN undefined_object"
        " THEN RAISE NOTICE '%, skipping', SQLERRM USING ERRCODE = SQLSTATE; "
        "END "
        "$$"
    )
    if echo:
        print(f"{status_msg} --> {revoke_statement}")
    if not dry_run:
        session.execute(sql_statement)


def _revoke_all_privileges_from_read_and_write_roles(
    session, pg_schema, dataset_name=None, echo=True, dry_run=False
):
    """Revoke all privileges that may have been previously granted to the scope_{} and write_{}
    roles. If dataset_name is provided, revoke only rights to the tables belonging to
    dataset_name.
    """

    status_msg = "Skipped" if dry_run else "Executed"
    # with engine.begin() as connection:
    result = session.execute(
        text(
            r"""
            SELECT rolname
            FROM pg_roles
            WHERE rolname LIKE 'scope\_%'
               OR rolname LIKE 'write\_%'
            """
        )
    )
    for rolname in result:
        if dataset_name:
            # for a single dataset
            revoke_statements = []
            for table in dataset_name.tables:
                revoke_statements.append(
                    f"REVOKE ALL PRIVILEGES ON {pg_schema}.{table.db_name()} FROM {rolname[0]}"
                )
            revoke_statement = ";".join(revoke_statements)
        else:
            revoke_statement = (
                f"REVOKE ALL PRIVILEGES ON ALL TABLES IN SCHEMA {pg_schema} FROM {rolname[0]};"
            )
        if echo:
            print(f"{status_msg} --> {revoke_statement}")
        if not dry_run:
            session.execute(text(revoke_statement))  # .execution_options(autocommit=True))


def _execute_grant(session, grant_statement, echo=True, dry_run=False):
    """Wrap the grant statement in an anonymous code block to catch reasonable exceptions.
    We don't want to break out of the session just because a table, column, or user doesn't
    exist.
    """

    status_msg = "Skipped" if dry_run else "Executed"
    sql_statement = f"""
        DO
        $$
            BEGIN
                {grant_statement};
            EXCEPTION
                WHEN undefined_table OR undefined_column OR undefined_object
                    THEN RAISE NOTICE '%, skipping', SQLERRM USING ERRCODE = SQLSTATE;
            END
        $$
        """
    if echo:
        print(f"{status_msg} --> {grant_statement}")
    if not dry_run:
        session.execute(sql_statement)


def _create_role_if_not_exists(session, role, echo=True, dry_run=False):
    """Wrap the create role statement in an anonymous code block to be able to catch exception
    Don't break out of the session just because the role already exists
    """

    status_msg = "Skipped" if dry_run else "Executed"
    create_role_statement = f"CREATE ROLE {role}"
    if role not in existing_roles:
        sql_statement = (
            "DO $$ "
            "BEGIN "
            f"{create_role_statement}; "
            "EXCEPTION"
            " WHEN duplicate_object"
            " THEN RAISE NOTICE '%, skipping', SQLERRM USING ERRCODE = SQLSTATE; "
            "END "
            "$$"
        )
        if echo:
            print(f"{status_msg} --> {create_role_statement}")
        if not dry_run:
            session.execute(text(sql_statement))
        existing_roles.add(role)


def scope_to_role(scope):
    return f"scope_{scope.lower().replace('/', '_')}"



from typing import List, Optional

from django.conf import settings
from django.core.management import BaseCommand
from django.db.models import Q

from schematools.contrib.django.models import Dataset
from schematools.types import DatasetSchema
from schematools.utils import dataset_paths_from_url, dataset_schemas_from_url, to_snake_case

from .create_tables import create_tables


class Command(BaseCommand):
    help = """Import Amsterdam schema files.

    import_schemas imports all schemas from $SCHEMA_URL,
    unless schema files/URLs are given as positional arguments.
    """
    requires_system_checks = False

    def add_arguments(self, parser):
        parser.add_argument("schema", nargs="*", help="Local schema files to import")
        parser.add_argument(
            "--schema-url",
            default=settings.SCHEMA_URL,
            help=f"Schema URL (default: {settings.SCHEMA_URL})",
        )
        parser.add_argument("--create-tables", dest="create_tables", action="store_true")
        parser.add_argument("--no-create-tables", dest="create_tables", action="store_false")
        parser.set_defaults(create_tables=False)

    def handle(self, *args, **options):
        if options["schema"]:
            datasets = self.import_from_files(options["schema"])
        else:
            datasets = self.import_from_url(options["schema_url"])

        if not datasets:
            self.stdout.write("No new datasets imported")
            return

        # Reasons for not creating tables directly are to manually configure the
        # "Datasets" model flags first. E.g. disable "enable_db", set a remote URL.
        if options["create_tables"]:
            create_tables(self, datasets, allow_unmanaged=True)

    def import_from_files(self, schema_files) -> List[Dataset]:
        """Import all schema definitions from the given files."""
        datasets = []
        for filename in schema_files:
            self.stdout.write(f"Loading schema from {filename}")
            schema = DatasetSchema.from_file(filename)
            dataset = self.import_schema(schema.id, schema)
            datasets.append(dataset)

        return datasets

    def import_from_url(self, schema_url) -> List[Dataset]:
        """Import all schema definitions from an URL"""
        self.stdout.write(f"Loading schema from {schema_url}")
        datasets = []

        schemas = dataset_schemas_from_url(schema_url)
        paths = dataset_paths_from_url(schema_url)
        for name, schema in schemas.items():
            self.stdout.write(f"* Processing {name}")
            dataset = self.import_schema(name, schema, paths[name])
            if dataset is not None:
                datasets.append(dataset)

        return datasets

    def import_schema(
        self, name: str, schema: DatasetSchema, path: Optional[str] = None
    ) -> Optional[Dataset]:
        """Import a single dataset schema."""

        created = False
        try:
            dataset = Dataset.objects.get(name=Dataset.name_from_schema(schema))
        except Dataset.DoesNotExist:
            try:
                # try getting default dataset by name and version
                dataset = Dataset.objects.filter(Q(version=None) | Q(version=schema.version)).get(
                    name=to_snake_case(schema.id)
                )
            except Dataset.DoesNotExist:
                # Give up, Create new dataset
                dataset = Dataset.create_for_schema(schema, path)
                created = True

        if created:
            self.stdout.write(f"  Created {name}")
            return dataset
        else:
            self.stdout.write(f"  Updated {name}")

            if dataset.is_default_version != schema.is_default_version:
                self.update_dataset_version(dataset, schema)

            updated = dataset.save_for_schema(schema)
            if path is not None:
                dataset.save_path(path)
            if updated:
                return dataset

        return None

    def update_dataset_version(self, dataset: Dataset, schema: DatasetSchema) -> Dataset:
        """
        Perform dataset version update, including changes to dataset tables.
        """
        if not dataset.is_default_version:
            # Dataset is currently not default. Can not be safely renamed.
            if schema.is_default_version:
                # Dataset is promoted to default. We need to rename current default,
                #  if it was not done yet.
                try:
                    current_default = Dataset.objects.get(name=Dataset.name_from_schema(schema))
                except Dataset.DoesNotExist:
                    pass
                else:
                    # Update current default dataset name to expected name.
                    if current_default.version:
                        current_default.name = to_snake_case(
                            f"{schema.id}_{current_default.version}"
                        )

                        current_default.save()

        dataset.name = Dataset.name_from_schema(schema)
        dataset.is_default_version = schema.is_default_version



from sqlalchemy import Column, DateTime, ForeignKey, ForeignKeyConstraint, String
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import relationship

__all__ = ["Base", "Dataset", "Table", "Field"]

Base = declarative_base()


class Dataset(Base):
    __tablename__ = "ams_schema_dataset"
    id = Column(String, primary_key=True)
    path = Column(String)
    type = Column(String)
    title = Column(String)
    description = Column(String)
    date_created = Column(DateTime)
    auth = Column(String)
    date_modified = Column(DateTime)
    license = Column(String)
    homepage = Column(String)
    language = Column(String)
    status = Column(String)
    version = Column(String)
    objective = Column(String)
    temporal_unit = Column(String)
    spatial = Column(String)
    legal_basis = Column(String)
    contact_point = Column(String)  # struct with name/email
    accrual_periodicity = Column(String)
    spatial_description = Column(String)
    spatial_coordinates = Column(String)
    theme = Column(String)
    publisher = Column(String)
    owner = Column(String)
    authorization_grantor = Column(String)
    keywords = Column(String)
    has_beginning = Column(DateTime)
    has_end = Column(DateTime)
    crs = Column(String)


class Table(Base):
    __tablename__ = "ams_schema_table"
    id = Column(String, primary_key=True)
    dataset_id = Column(String, ForeignKey("ams_schema_dataset.id"), primary_key=True)
    type = Column(String)
    title = Column(String)
    description = Column(String)
    display = Column(String)
    required = Column(String)
    date_created = Column(DateTime)
    auth = Column(String)
    date_modified = Column(DateTime)
    license = Column(String)
    schema = Column(String)
    schema_version = Column(String)

    dataset = relationship("Dataset", backref="tables")


class Field(Base):
    __tablename__ = "ams_schema_field"
    name = Column(String, primary_key=True)
    table_id = Column(String, primary_key=True)
    ref = Column(String)
    dataset_id = Column(String)
    type = Column(String)
    title = Column(String)
    description = Column(String)
    format = Column(String)
    uri = Column(String)
    relation = Column(String)
    auth = Column(String)
    unit = Column(String)
    enum = Column(String)

    table = relationship("Table", backref="fields")

    __table_args__ = (
        ForeignKeyConstraint(
            ["table_id", "dataset_id"],
            ["ams_schema_table.id", "ams_schema_table.dataset_id"],
        ),
    )
